from PIL import Image
from .. import requirePackage


def is_svg(imgPath: str):
    """
    判断图片是否为svg格式

    check if the image is svg

    :param imgPath: 图片路径 | image path
    :return:
    """
    return imgPath.endswith(".svg") or imgPath.endswith('.svgz')


def is_eps(imgPath: str):
    """
    判断图片是否为eps格式

    check if the image is eps

    :param imgPath: 图片路径 | image path
    :return:
    """
    return imgPath.endswith(".eps") or imgPath.endswith('.epsi') or imgPath.endswith('.epsf')


def is_pdf(imgPath: str):
    """
    判断图片是否为pdf格式

    check if the image is pdf

    :param imgPath: 图片路径 | image path
    :return:
    """
    return imgPath.endswith(".pdf")


def is_heic(imgPath: str):
    """
    判断图片是否为heic格式

    check if the image is heic

    :param imgPath: 图片路径 | image path
    :return:
    """
    return imgPath.endswith(".heic")


def is_pdf(imgPath: str):
    """
    判断图片是否为pdf格式

    check if the image is pdf

    :param imgPath: 图片路径 | image path
    :return:
    """
    return imgPath.endswith(".pdf")


def topng(imgPath: str):
    """
    将图片路径所指的图片转为jpg格式

    transform image to jpg

    :param imgPath: 图片路径 | image path
    :return:
    """
    if is_svg(imgPath):
        requirePackage("cairosvg", "svg2png")(
            url=imgPath, write_to=imgPath + ".png", dpi=300
        )
    elif is_eps(imgPath):
        requirePackage("wand.image", "Image")(filename=imgPath).save(filename=imgPath + ".png")
    elif is_heic(imgPath):
        requirePackage("pillow_heif", "read_heif")(imgPath).save(imgPath + ".png")
    elif is_pdf(imgPath):
        requirePackage('pdfplumber', 'open')(imgPath).pages[0].to_image(resolution=300).original.save(imgPath + ".png")
    else:
        Image.open(imgPath).save(imgPath + ".png", quality=100, optimize=True)


def tojpg(imgPath: str):
    """
    将图片路径所指的图片转为jpg格式

    transform image to jpg

    :param imgPath: 图片路径 | image path
    :return:
    """
    if is_svg(imgPath):
        imgPath = requirePackage("io", "BytesIO", "io")(
            requirePackage("cairosvg", "svg2png", "cairosvg")(url=imgPath, dpi=300)
        )
    elif is_eps(imgPath):
        imgPath = requirePackage("io", "BytesIO", "io")(
            requirePackage("wand.image", "Image", "wand.image")(filename=imgPath).make_blob()
        )
    elif is_heic(imgPath):
        imgPath = requirePackage("io", "BytesIO", "io")(
            requirePackage("pillow_heif", "read_heif")(imgPath).save_to_memory()
        )
    elif is_pdf(imgPath):
        imgPath = requirePackage('pdfplumber', 'open')(imgPath).pages[0].to_image(resolution=300).original.save(imgPath + ".jpg")
    else:
        Image.open(imgPath).convert("RGB").save(imgPath + ".jpg", quality=100, optimize=True)


def topdf(imgPath: str):
    """
    将图片路径所指的图片转为pdf格式

    transform image to pdf

    :param imgPath: 图片路径 | image path
    :return:
    """
    if is_svg(imgPath):
        requirePackage("cairosvg", "svg2pdf")(
            url=imgPath, write_to=imgPath + ".pdf", dpi=300
        )
    elif is_eps(imgPath):
        requirePackage("wand.image", "Image")(filename=imgPath).save(filename=imgPath + ".pdf")
    else:
        Image.open(imgPath).save(imgPath + ".pdf", quality=100, optimize=True)

def imgsConcat(imgs: list):
    """
    合并图片
    """

    from io import BytesIO
    from .. import (
        requirePackage,
        qs_default_console,
        qs_default_status,
        qs_error_string,
        qs_config,
    )

    Image = requirePackage("PIL", "Image", "Pillow")
    try:
        imgs = [Image.open(BytesIO(i)) for i in imgs if i]
    except:
        qs_default_console.print(qs_error_string, "样品图获取失败!")
        return

    if not imgs:
        qs_default_console.print(qs_error_string, "无样品图")
        return

    with qs_default_status("拼接图片中") as st:
        import array, fcntl, termios, sys
        buf = array.array('H', [0, 0, 0, 0])
        fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, buf)
        width, max_height = buf[2], buf[3]
        one_width = width
        heights_len = min(len(imgs), 3)
        heights = [0] * heights_len
        for i in imgs:
            one_height = int(one_width * i.size[1] / i.size[0])
            heights[heights.index(min(heights))] += one_height

        st.update("嗅探最佳拼接方式")

        while max(heights) > max_height and heights_len < len(imgs):
            heights_len += 1
            heights = [0] * heights_len
            one_width = int(width / heights_len)
            for i in imgs:
                one_height = int(one_width * i.size[1] / i.size[0])
                heights[heights.index(min(heights))] += one_height

        result = Image.new("RGBA", (one_width * heights_len * heights_len, max(heights) * heights_len))
        heights = [0] * heights_len

        imgs = [
            i.resize((one_width * heights_len, int(one_width * i.size[1] / i.size[0]) * heights_len)) for i in imgs
        ]
        imgs = sorted(imgs, key=lambda i: -i.size[0] * i.size[1])

        for i in imgs:
            min_height_index = heights.index(min(heights))
            result.paste(i, (one_width * heights_len * min_height_index, heights[min_height_index]))
            heights[min_height_index] += i.size[1]
    return result



# coding=utf-8
"""
柱状图

histogram & Bar
"""


def DataTransformBar(has_size_info: bool = True):
    """
    创建一个数据传输进度条

    Create a data transfer progress bar

    :param has_size_info: 是否有总任务量
    :return: rich.progress.Progress
    """
    from rich.progress import (
        Progress,
        TextColumn,
        BarColumn,
        DownloadColumn,
        TransferSpeedColumn,
        TimeRemainingColumn,
    )
    from .. import qs_default_console

    if has_size_info:
        return Progress(
            TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(binary_units=True),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
            console=qs_default_console,
        )
    else:
        from .. import user_lang

        return Progress(
            TextColumn(
                "[bold blue]{task.fields[filename]} [red]"
                + ("Unknow size" if user_lang != "zh" else "未知大小"),
                justify="right",
            ),
            BarColumn(bar_width=None),
            DownloadColumn(),
            console=qs_default_console,
        )


def NormalProgressBar(task_name, total):
    from .. import qs_default_console
    from rich.progress import Progress

    progress = Progress()
    task_id = progress.add_task(task_name, total=total, console=qs_default_console)
    return progress, task_id



from QuickProject import QproDefaultConsole as qs_default_console
from QuickProject import _ask, user_lang
from langsrc import LanguageDetector
import json
import sys
import os

is_iterm2 = "ITERM_SESSION_ID" in os.environ
is_kitty = "KITTY_PID" in os.environ or "GHOSTTY_BIN_DIR" in os.environ

platform = sys.platform
if platform.startswith("win"):
    dir_char = "\\"
else:
    dir_char = "/"


translate_engines = ["default", "TencentCloud", "DeepL", "DeepLX", "AITranslate"]
lang_detector = LanguageDetector(user_lang, os.path.join(os.path.dirname(__file__), "lang.json"))

questions = {
    "default_translate_engine": {
        "type": "list",
        "message": lang_detector['ask_translate_engine'],
        "choices": translate_engines,
        "default": "default",
    },
    "terminal_app": {
        "type": "list",
        "name": "force_show_img",
        "message": lang_detector['ask_force_show_img'],
        "choices": ["iTerm2", "Kitty Series"],
        "default": "iTerm2" if is_iterm2 else "Kitty Series",
    },
    "default_proxy": {
        "type": "input",
        "name": "default_proxy",
        "message": lang_detector['ask_proxy'],
        "default": "Not set | 暂不设置",
    },
    "gpt": {
        "type": "list",
        "message": lang_detector['ask_gpt'],
        "choices": ["openai", "poe", "alapi"],
    },
    "gpt-model": {
        "type": "input",
        "message": lang_detector['ask_gpt_model']
    },
    "gpt-api": {
        "type": "input",
        "message": lang_detector['ask_gpt_api']
    },
    "gpt-url": {
        "type": "input",
        "message": lang_detector['ask_gpt_url']
    },
}


class QsConfig:
    def __init__(self, configPath):
        self.path = configPath
        if os.path.exists(configPath):
            try:
                with open(configPath, "r") as f:
                    self.config = json.loads(f.read())
            except:
                with open(configPath, "r", encoding="utf8") as f:
                    self.config = json.loads(f.read(), encoding="utf8")
        else:
            self.config = json.loads(
                """{
  "basic_settings": {
    "default_translate_engine": {
      "index": 0,
      "support": ["default", "TencentCloud", "DeepL", "DeepLX", "AITranslate"]
    },
    "default_proxy": "user:password@ip:port or ip:port",
    "gpt": {
        "index": "openai",
        "support": {
            "openai": {
                "API_KEY": "",
                "API_URL": ""
            },
            "poe": "",
            "alapi": ""
        },
        "model": "",
        "translate": "",
        "prompt": {
            "General": "You are a powerful assistant with knowledge spanning various fields."
        }
    }
  },
  "API_settings": {
    "rmbg": "GET: https://www.remove.bg",
    "smms": "GET: https://sm.ms",
    "darksky": "GET: https://darksky.net/",
    "aliyun_oss_acid": "GET: https://www.aliyun.com/product/oss",
    "aliyun_oss_ackey": "GET: https://www.aliyun.com/product/oss",
    "aliyun_oss_bucket_url": "GET: https://www.aliyun.com/product/oss",
    "aliyun_oss_df_bucket": "GET: https://www.aliyun.com/product/oss",
    "txyun_scid": "GET: https://console.cloud.tencent.com/",
    "txyun_sckey": "GET: https://console.cloud.tencent.com/",
    "txyun_cos_df_bucket": "GET: https://console.cloud.tencent.com/",
    "txyun_df_region": "GET: ap-[location]",
    "qiniu_ac_key": "GET: http://qiniu.com/",
    "qiniu_sc_key": "GET: http://qiniu.com/",
    "qiniu_bk_name": "GET: [Qiniu Bucket Name]",
    "AipImageAPP_ID": "GET: https://cloud.baidu.com/product/imageprocess",
    "AipImageAPP_KEY": "GET: https://cloud.baidu.com/product/imageprocess",
    "AipImageSECRET_KEY": "GET: https://cloud.baidu.com/product/imageprocess",
    "AipNlpAPP_ID" : "GET: https://cloud.baidu.com/product/nlp_apply",
    "AipNlpAPP_KEY": "GET: https://cloud.baidu.com/product/nlp_apply",
    "AipNlpSECRET_KEY": "GET: https://cloud.baidu.com/product/nlp_apply",
    "alapi_token": "GET: https://user.alapi.cn/",
    "lolicon_token": "GET: https://api.lolicon.app/#/setu?id=apikey",
    "DeepL": "GET: https://www.deepl.com/zh/pro-api?cta=header-pro-api/",
    "DeepLX": "GET: https://github.com/OwO-Network/DeepLX"
  }
}"""
            )
            self.config["basic_settings"]["default_translate_engine"][
                "index"
            ] = translate_engines.index(_ask(questions["default_translate_engine"]))
            self.config["basic_settings"]["default_proxy"] = _ask(
                questions["default_proxy"]
            )
            self.update()
            qs_default_console.print(
                "\n" + lang_detector['config_store_in'],
                f"[bold green]{configPath}[/bold green]",
            )
            _ask(
                {
                    "type": "confirm",
                    "message": lang_detector['confirm'],
                    "default": True,
                }
            )

    def update(self):
        with open(self.path, "w") as f:
            json.dump(self.config, f, indent=4, separators=(",", ": "))

    def basicSelect(self, key: str, default=None):
        if key not in self.config["basic_settings"]:
            if default is not None:
                self.config["basic_settings"][key] = default
                self.update()
            elif key == 'gpt':
                _info = {
                    "index": _ask(questions['gpt']),
                    "support": {
                        "openai": {
                            "API_KEY": "",
                            "API_URL": ""
                        },
                        "poe": "",
                        "alapi": ""
                    },
                    "model": _ask(questions['gpt-model'])
                }
                if _info['index'] == 'openai':
                    _info['support']['openai']['API_KEY'] = _ask(questions['gpt-api'])
                    _info['support']['openai']['API_URL'] = _ask(questions['gpt-url'])
                else:
                    _info[_info['index']] = _ask(questions['gpt-api'])
                self.config["basic_settings"][key] = _info
                self.update()
            else:
                self.config["basic_settings"][key] = _ask(questions[key])
                self.update()
        return self.config["basic_settings"][key]

    def apiSelect(self, key):
        if key not in self.config["API_settings"]:
            raise KeyError
        return self.config["API_settings"].get(key, None)

    def basicUpdate(self, key: str, value):
        self.config["basic_settings"][key] = value
        return self.update()

    def apiUpdate(self, key: str, value: str):
        self.config["API_settings"][key] = value
        return self.update()



# coding=utf-8
"""
调用各种系统工具

Call various system tools
"""
import QuickStart_Rhy.Wrapper as _wrapper

miss_file = [".DS_Store"]


def __latest_filename(name):
    import os

    cur = os.getcwd()
    while cur != os.path.dirname(cur):
        if os.path.exists(os.path.join(cur, name)):
            return os.path.join(cur, name)
        cur = os.path.dirname(cur)
    return os.path.join(cur, name) if os.path.exists(os.path.join(cur, name)) else ""


def top():
    """
    CPU和内存监测

    CPU and memory monitoring

    :return: None
    """
    import sys

    sys.argv = ["bpytop"] + sys.argv[2:]

    from . import requirePackage

    requirePackage("bpytop", "main")()


def clear_mem():
    """
    清理系统内存

    Clean system memory

    :return: None
    """
    from .SystemTools import clear_mem

    clear_mem()


def go_github():
    """
    自动识别当前文件夹.git/config中的地址，并通过浏览器打开

    Automatically recognize the address in the current folder .git/config and open it through a browser
    """
    import os
    from . import (
        qs_default_console,
        qs_error_string,
        user_lang,
        open_url,
        requirePackage,
    )

    config_path = __latest_filename(".git/config")
    if not os.path.exists(config_path):
        qs_default_console.print(
            qs_error_string,
            "No a git dictionary" if user_lang != "zh" else "不是 git 文件夹",
        )
        return
    config = requirePackage("configparser", "ConfigParser")()
    config.read(filenames=config_path)
    url_ls = []
    for section in config.sections():
        if section.startswith("remote"):
            url_ls.append(config[section]["url"].replace(".git", ""))
    open_url(url_ls)


def mktar():
    """
    创建tar包

    Create a tar packages

    :return: None
    """

    @_wrapper.mkCompressPackageWrap
    def _mktar(file_path: str = ""):
        from .SystemTools.Compress import Tar

        return Tar(file_path + ".tar.gz", "w")

    return _mktar()


def untar():
    """
    解压tar包

    Unpack the tar packages

    :return: None
    """

    @_wrapper.unCompressPackageWrap
    def _untar(file_path: str = ""):
        from .SystemTools.Compress import Tar

        return Tar(file_path)

    return _untar()


def mkzip():
    """
    创建ZIP包

    Create a ZIP package

    :return: None
    """

    @_wrapper.mkCompressPackageWrap
    def _mkzip(file_path: str = ""):
        from .SystemTools.Compress import Zip

        return Zip(file_path + ".zip", "w")

    return _mkzip()


def unzip():
    """
    解压ZIP包

    Unpack the ZIP package

    :return: None
    """

    @_wrapper.unCompressPackageWrap
    def _unzip(file_path: str = ""):
        from .SystemTools.Compress import Zip

        return Zip(file_path, "r")

    return _unzip()


def unrar():
    """
    解压RAR包

    Extract RAR package

    :return: None
    """

    @_wrapper.unCompressPackageWrap
    def _unrar(file_path: str = ""):
        from .SystemTools.Compress import Rar

        return Rar(file_path)

    return _unrar()


def mk7z():
    """
    创建7z包

    Create 7z package

    :return: None
    """

    @_wrapper.mkCompressPackageWrap
    def _mk7z(file_path: str = ""):
        from .SystemTools.Compress import SevenZip

        return SevenZip(file_path + ".7z", "w")

    return _mk7z()


def un7z():
    """
    解压7z包

    Extract 7z package

    :return:
    """

    @_wrapper.unCompressPackageWrap
    def _un7z(file_path: str = ""):
        from .SystemTools.Compress import SevenZip

        return SevenZip(file_path)

    return _un7z()


@_wrapper.HashWrapper()
def md5():
    """
    获取文件md5值
    :return:
    """


@_wrapper.HashWrapper()
def sha1():
    """
    获取文件sha1值
    :return:
    """


@_wrapper.HashWrapper()
def sha256():
    """
    获取文件sha256值
    :return:
    """


@_wrapper.HashWrapper()
def sha512():
    """
    获取文件sha512值
    :return:
    """


def diff_dir():
    """
    对比两个文件夹差异，并生成相应html对比结果
    :return:
    """
    from . import user_lang, qs_default_console, qs_info_string, qs_default_status
    from .SystemTools.Diff import DictionaryFiles
    import sys

    if "-h" in sys.argv:
        qs_default_console.print(
            qs_info_string, "Usage: qs diff <dir1> <dir2> [-x <name or regex pattern>]"
        )

    d1, d2 = sys.argv[2:4]
    apply_ignore = sys.argv[sys.argv.index("-x") + 1 :] if "-x" in sys.argv else None
    d1 = DictionaryFiles(d1, apply_ignore)
    d2 = DictionaryFiles(d2, apply_ignore)

    if not (d1.available and d2.available):
        return

    with qs_default_status(
        "Generating diff result.." if user_lang != "zh" else "生成对比结果中.."
    ):
        from .SystemTools.Diff import DiffFilesToStructHtml

        DiffFilesToStructHtml(d1, d2).generate()


def mount_dmg():
    """
    挂载镜像 

    :return:
    """
    from .SystemTools.DiskMac import DMG
    import sys

    DMG().mount(sys.argv[2])


def unmount_dmg():
    """
    卸载镜像
    """
    from .SystemTools.DiskMac import DMG
    from .TuiTools.Table import qs_default_table
    from . import qs_default_console, user_lang, qs_info_string, _ask

    disks = DMG()
    _ls = disks.get_disk_list()[2:]
    if not _ls:
        return qs_default_console.print(
            qs_info_string, "No DMG disk found" if user_lang != "zh" else "没有找到 dmg 磁盘"
        )
    table = qs_default_table(["Disk", "Type", "Size"])
    for disk in _ls:
        table.add_row(disk["path"], disk["type"], disk["size"])
    qs_default_console.print(table, justify="center")
    disk_paths = _ask(
        {
            "type": "checkbox",
            "message": "Select a disk to umount" if user_lang != "zh" else "选择要卸载的磁盘",
            "choices": [{'name': disk["path"], 'checked': True} for disk in _ls],
        }
    )
    for disk_path in disk_paths:
        disks.unmount(disk_path)



# coding=utf-8
"""
曲线图

Line Graph
"""
import io
from .. import requirePackage

diagram = requirePackage("diagram")


class Line:
    def __init__(self, values: list, height: int, width: int):
        """
        折线图

        line chart

        :param values: 值列表
        :param height: 高度
        :param width: 宽度
        """
        self.values = values
        self.width = height
        self.height = width

    def __str__(self):
        """
        绘图并返回字符串

        Draw and return a string

        :return: 折线图的字符串
        """
        opt = diagram.DOption()
        opt.size = diagram.Point([self.height, self.width])
        opt.mode = "g"
        stream = io.BytesIO()
        gram = diagram.DGWrapper(
            data=[self.values, range(len(self.values))], dg_option=opt, ostream=stream
        )
        gram.show()
        return str(stream.getvalue(), encoding="utf-8")

    def push(self, x: float):
        """
        尾部添加值

        Tail add value

        :param x: 值
        :return: None
        """
        self.values.append(x)

    def pop(self):
        """
        删除首部值

        Delete bow value

        :return: None
        """
        self.values.pop(0)



"""
DeepL 翻译 API
"""

from . import pre_check
from .. import user_lang, requirePackage


_translator = requirePackage("deepl", "Translator")(pre_check("DeepL"))  # 持久翻译复用


def translate(text, target_lang=user_lang.lower()):
    """翻译文本"""
    if target_lang == 'en':
        target_lang = 'en-us'
    return _translator.translate_text(text, target_lang=target_lang).text


if __name__ == "__main__":
    print(translate("Hello World!"))



# coding=utf-8
"""
腾讯云相关API

Tencent cloud API
"""
import json
from .. import requirePackage, user_lang
from . import pre_check


class TxCOS:
    def __init__(self, scid=None, sckey=None, region=None, bucket=None, cdn_url=None):
        """
        初始化并登陆腾讯云对象存储

        Initialize and log in Tencent Cloud object storage

        :param scid: SecretId
        :param sckey: SecretKey
        :param region: 地区
        :param bucket: 桶名称
        :param cdn_url: CDN地址

        :return: None
        """
        if not (scid and sckey and region and bucket):
            scid, sckey, self.region, self.df_bucket = pre_check(
                "txyun_scid", "txyun_sckey", "txyun_df_region", "txyun_cos_df_bucket"
            )
        self.cdn_url = (
            pre_check("txyun_df_cdn_url", ext=False) if not cdn_url else cdn_url
        )
        if self.cdn_url and not self.cdn_url.endswith("/"):
            self.cdn_url += "/"

        qcloud_cos = requirePackage("qcloud_cos", real_name="cos-python-sdk-v5")

        config = qcloud_cos.CosConfig(
            Region=self.region, SecretId=scid, SecretKey=sckey
        )
        self.client = qcloud_cos.CosS3Client(config)

    def get_func_table(self):
        """
        获取对象支持的函数表

        Gets a table of functions supported by the object

        :return: 函数表
        """
        return {
            "-up": self.upload,
            "-rm": self.remove,
            "-dl": self.download,
            "-ls": self.list_bucket,
        }

    def upload(self, filePath: str, bucket: str = None, key: str = None):
        """
        上传文件

        Upload file

        :param key: 在COS上的文件路径
        :param filePath: 文件地址
        :param bucket: 桶名称，缺省使用self.df_bucket
        :return: None
        """
        from .. import dir_char

        bucket = bucket if bucket else self.df_bucket
        filename = filePath.strip()
        self.client.upload_file(
            Bucket=bucket,
            LocalFilePath=filename,
            Key=key if key else filename.replace(dir_char, "/"),
        )

    def download(self, filename: str, bucket: str = None):
        """
        下载文件

        Download file

        :param filename: 文件在桶中的路径
        :param bucket: 桶名称，缺省使用self.df_bucket
        :return: None
        """
        from ..NetTools.NormalDL import normal_dl

        bucket = bucket if bucket else self.df_bucket

        url = (
            self.cdn_url + filename
            if bucket == self.df_bucket and self.cdn_url
            else "https://"
            + bucket
            + ".cos."
            + self.region
            + ".myqcloud.com/"
            + filename
        )
        normal_dl(url)  # * 由于腾讯云没有提供可调用的下载sdk，因此使用qs下载引擎下载

    def remove(self, filePath: str, bucket: str = None):
        """
        删除桶中的文件

        Delete files in the bucket

        :param filePath: 文件在桶中的路径
        :param bucket: 桶名称，缺省使用self.df_bucket
        :return: None
        """
        bucket = bucket if bucket else self.df_bucket
        self.client.delete_object(bucket, filePath)

    def list_bucket(self, bucket: str = None):
        """
        展示桶中的全部文件信息

        Displays all file information in the bucket

        :param bucket: 桶名称，缺省使用self.df_bucket
        :return: None
        """
        from .. import qs_default_console, qs_info_string
        from ..NetTools.NormalDL import size_format
        from ..TuiTools.Table import qs_default_table
        from rich.text import Text

        bucket = bucket if bucket else self.df_bucket
        ls = self.client.list_objects(Bucket=bucket)["Contents"]
        tb = qs_default_table(
            ["File", "Size"] if user_lang != "zh" else ["文件", "体积"], title="Tencnet COS"
        )
        for obj in ls:
            tb.add_row(
                Text(obj["Key"], justify="left"),
                Text(size_format(int(obj["Size"]), align=True), justify="right"),
            )
        qs_default_console.print(
            qs_info_string,
            "Bucket Url:" if user_lang != "zh" else "桶链接:",
            self.cdn_url
            if bucket == self.df_bucket and self.cdn_url
            else "https://" + bucket + ".cos." + self.region + ".myqcloud.com/",
        )
        qs_default_console.print(tb, justify="center")


class Translate:

    def __init__(self, scid, sckey, region):
        """
        初始化并登陆腾讯翻译接口

        Initialize and log in Tencent translation interface

        :param scid: secret id
        :param sckey: secret key
        :param region: 地区
        """
        if not (scid and sckey and region):
            scid, sckey, region = pre_check(
                "txyun_scid", "txyun_sckey", "txyun_df_region"
            )
        requirePackage("langdetect", "DetectorFactory").seed = 0
        cred = requirePackage(
            "tencentcloud.common", "credential", real_name="tencentcloud-sdk-python"
        )
        http_profile = requirePackage(
            "tencentcloud.common.profile.http_profile",
            "HttpProfile",
            real_name="tencentcloud-sdk-python",
        )()
        http_profile.endpoint = "tmt.tencentcloudapi.com"
        clientProfile = requirePackage(
            "tencentcloud.common.profile.client_profile",
            "ClientProfile",
            real_name="tencentcloud-sdk-python",
        )()
        clientProfile.httpProfile = http_profile
        self.client = requirePackage(
            "tencentcloud.tmt.v20180321.tmt_client",
            "TmtClient",
            real_name="tencentcloud-sdk-python",
        )(cred, region, clientProfile)

    @staticmethod
    def langdetect(text: str) -> str:
        """
        获取文本的语言类型

        Gets the language type of the text

        :param text: 待识别文本
        :return: 语言类型
        """
        return requirePackage("langid", "classify")(text)[0]

    def translate(
        self, text: str, from_lang: str = None, target_lang: str = user_lang
    ) -> str:
        """
        翻译文本至默认语言

        Translate text to the default language

        :param from_lang: 文本语言
        :param target_lang: 目标语言
        :param text: 文本
        :return: 翻译结果
        """
        req = requirePackage(
            "tencentcloud.tmt.v20180321.models", "TextTranslateRequest"
        )()
        req.from_json_string(
            json.dumps(
                {
                    "SourceText": text,
                    "Source": self.langdetect(text) if not from_lang else from_lang,
                    "Target": target_lang,
                    "ProjectId": 0,
                }
            )
        )
        return json.loads(self.client.TextTranslate(req).to_json_string())["TargetText"]


def translate(text: str, from_lang: str = None, target_lang: str = user_lang) -> str:
    return Translate().translate(text, from_lang, target_lang)



# coding=utf-8
"""
qs用到的函数装饰器, 在使用它们前, 请确保你理解注释描述的内容

Before using function decorators used by QS, please make sure that you understand the content described by the annotation.
"""
from QuickProject.__config__ import set_timeout


def mkCompressPackageWrap(func):
    """
    创建压缩文件的通用函数装饰器, 被装饰的函数将被传递压缩文件名, 你需要返回:
        QuickStart_Rhy.SystemTools.Compress._NormalCompressedPackage
    的子类对象, 并将读写状态设为 写 状态

    General function decorator for creating compressed files, the decorated function will be passed the compressed file
    name, which you need to return:
        QuickStart_Rhy.SystemTools.Compress._NormalCompressedPackage
    's sub object, and set read-write mode as write

    :param func: func(filePath: str = '') -> QuickStart_Rhy.SystemTools.Compress._NormalCompressedPackage
    :return: 装饰器 | wrapper
    """

    def wrapper():
        from ..SystemTools.Compress import get_compress_package_name

        packages_name, ls = get_compress_package_name()
        packages = func(packages_name)

        for i in ls:
            packages.add_file(i)
        packages.save()

    return wrapper


def unCompressPackageWrap(func):
    """
    解压缩文件的通用函数装饰器, 被装饰的函数将被传递压缩文件名, 你需要返回:
        QuickStart_Rhy.SystemTools.Compress._NormalCompressedPackage
    的子类对象, 并将读写状态设为 读 状态

    General function decorator for extracting compressed files, the decorated function will be passed the compressed file
    name, which you need to return:
        QuickStart_Rhy.SystemTools.Compress._NormalCompressedPackage
    's sub object, and set read-write mode as read

    :param func: func(filePath: str = '') -> QuickStart_Rhy.SystemTools.Compress._NormalCompressedPackage
    :return: 装饰器 | wrapper
    """

    def wrapper():
        import os
        import sys

        file_names = sys.argv[2:]
        if not file_names:
            exit("No enough parameters")
        from .. import qs_default_console, qs_error_string
        from ..NetTools.NormalDL import core_num
        from ..ThreadTools import ThreadPoolExecutor, wait

        pool = ThreadPoolExecutor(max_workers=max(core_num // 2, 4))
        job_q = []

        def run(path):
            if os.path.exists(path):
                try:
                    cur_tar = func(path)
                    cur_tar.extract()
                except Exception as e:
                    qs_default_console.log(qs_error_string, repr(e))
            else:
                qs_default_console.log(
                    qs_error_string, "No such file or dictionary: %s" % path
                )

        for file_name in file_names:
            job_q.append(pool.submit(run, file_name))
        wait(job_q)

    return wrapper


def HashWrapper():
    """
    文件哈希值计算(通用函数)
    :return:
    """

    def Wrapper(func):
        algorithm = func.__name__

        def _wrapper():
            import sys, os
            from .. import qs_default_console, qs_info_string, requirePackage

            ls = sys.argv[2:]
            if not ls:
                qs_default_console.print(
                    qs_info_string, "Usage: qs %s file1 file2 ..." % algorithm
                )
                return

            func = requirePackage(".SystemTools.FileHash", algorithm)
            for file in ls:
                qs_default_console.print(
                    f"\[[bold magenta]{algorithm}[/]]",
                    f"[underline]{os.path.basename(file)}[/]: '{func(file)}'",
                )

        return _wrapper

    return Wrapper



"""
在终端预览图片 | 仅有MacOS下的iTerm2可用

preview image on terminal | Only iTerm2 is available,
"""
from .. import qs_default_console, qs_default_status
from .. import user_lang, qs_config
import math
import base64
import sys
import os
import struct
import io

from ..__config__ import is_iterm2, is_kitty

if not is_iterm2 and not is_kitty and (terminal_app := qs_config.basicSelect("terminal_app")):
    terminal_app = terminal_app.lower()
    if terminal_app == 'iterm2':
        is_iterm2 = True
    elif terminal_app == 'kitty series':
        is_kitty = True

TMUX_WRAP_ST = b"\033Ptmux;"
TMUX_WRAP_ED = b"\033\\"

OSC = b"\033]"
CSI = b"\033["
ST = b"\a"  # \a = ^G (bell)


def get_image_shape(buf):
    """
    Extracts image shape as 2-tuple (width, height) from the content buffer.

    Supports GIF, PNG and other image types (e.g. JPEG) if PIL/Pillow is installed.
    Returns (None, None) if it can't be identified.
    """

    def _unpack(fmt, buffer, mode="Image"):
        try:
            return struct.unpack(fmt, buffer)
        except struct.error:
            raise ValueError("Invalid {} file".format(mode))

    # TODO: handle 'stream-like' data efficiently, not storing all the content into memory
    L = len(buf)

    if L >= 10 and buf[:6] in (b"GIF87a", b"GIF89a"):
        return _unpack("<hh", buf[6:10], mode="GIF")
    elif L >= 24 and buf.startswith(b"\211PNG\r\n\032\n") and buf[12:16] == b"IHDR":
        return _unpack(">LL", buf[16:24], mode="PNG")
    elif L >= 16 and buf.startswith(b"\211PNG\r\n\032\n"):
        return _unpack(">LL", buf[8:16], mode="PNG")
    else:
        # everything else: get width/height from PIL
        # TODO: it might be inefficient to write again the memory-loaded content to buffer...
        b = io.BytesIO()
        b.write(buf)

        try:
            from .. import requirePackage
            im = requirePackage("PIL", "Image", "Pillow").open(b)
            return im.width, im.height
        except (IOError, OSError) as ex:
            # PIL.Image.open throws an error -- probably invalid byte input are given
            sys.stderr.write(
                "Warning: PIL cannot identify image; this may not be an image file"
                + "\n"
            )
        except ImportError:
            # PIL not available
            sys.stderr.write(
                "Warning: cannot determine the image size; please install Pillow" + "\n"
            )
            sys.stderr.flush()
        finally:
            b.close()

        return None, None


def _isinstance(obj, module, clsname):
    """A helper that works like isinstance(obj, module:clsname), but even when
    the module hasn't been imported or the type is not importable."""

    if module not in sys.modules:
        return False

    try:
        clstype = getattr(sys.modules[module], clsname)
        return isinstance(obj, clstype)
    except AttributeError:
        return False


def to_content_buf(data, fmt="png", width: int = 0, height: int = 0, bypass=False):
    # TODO: handle 'stream-like' data efficiently, rather than storing into RAM
    from .. import requirePackage
    if bypass:
        return data

    def pixels_exceed_limit(w, h) -> tuple:
        limit = 2097152
        if w * h > limit:
            downsize_rate = math.sqrt((w * h) / limit)
            w = math.floor(w / downsize_rate)
            h = math.floor(h / downsize_rate)
        return (w, h)
    
    if isinstance(data, bytes):
        # force transform data to png format
        img = requirePackage("PIL", "Image", "Pillow").open(io.BytesIO(data))
        if width and height:
            rate = img.size[0] / img.size[1]
            ass_height = width / rate
            ass_width = rate * height
            if ass_height > height:
                width = math.floor(height * rate)
            if ass_width > width:
                height = math.floor(width / rate)
            img = img.resize((width, height))
        else:
            w, h = pixels_exceed_limit(img.size[0], img.size[1])
            img = img.resize((w, h))
        with io.BytesIO() as buf:
            img.save(buf, format=fmt)
            return buf.getvalue()

    elif isinstance(data, io.BufferedReader):
        buf = data
        img = requirePackage("PIL", "Image", "Pillow").open(buf)
        if width and height:
            rate = img.size[0] / img.size[1]
            ass_height = width / rate
            ass_width = rate * height
            if ass_height > height:
                width = math.floor(height * rate)
            if ass_width > width:
                height = math.floor(width / rate)
            img = img.resize((width, height))
        else:
            width, height = pixels_exceed_limit(img.size[0], img.size[1])
            img = img.resize((width, height))
        with io.BytesIO() as buf:
            img.save(buf, format=fmt)
            return buf.getvalue()
        return buf.read()

    elif isinstance(data, io.TextIOWrapper):
        buf = data.buffer
        img = requirePackage("PIL", "Image", "Pillow").open(buf)
        if width and height:
            rate = img.size[0] / img.size[1]
            ass_height = width / rate
            ass_width = rate * height
            if ass_height > height:
                width = math.floor(height * rate)
            if ass_width > width:
                height = math.floor(width / rate)
            img = img.resize((width, height))
        else:
            width, height = pixels_exceed_limit(img.size[0], img.size[1])
            img = img.resize((width, height))
        with io.BytesIO() as buf:
            img.save(buf, format=fmt)
            return buf.getvalue()
        return buf.read()

    elif _isinstance(data, "numpy", "ndarray"):
        # numpy ndarray: convert to png
        im = data
        if len(im.shape) == 2:
            mode = "L"  # 8-bit pixels, grayscale
            im = im.astype(sys.modules["numpy"].uint8)
        elif len(im.shape) == 3 and im.shape[2] in (3, 4):
            mode = None  # RGB/RGBA
            if im.dtype.kind == "f":
                im = (im * 255).astype("uint8")
        else:
            raise ValueError(
                "Expected a 3D ndarray (RGB/RGBA image) or 2D (grayscale image), "
                "but given shape: {}".format(im.shape)
            )

        with io.BytesIO() as buf:
            # mode: https://pillow.readthedocs.io/en/4.2.x/handbook/concepts.html#concept-modes
            img = requirePackage("PIL", "Image", "Pillow").fromarray(im, mode=mode)
            if width and height:
                rate = img.size[0] / img.size[1]
                ass_height = width / rate
                ass_width = rate * height
                if ass_height > height:
                    width = math.floor(height * rate)
                if ass_width > width:
                    height = math.floor(width / rate)
                img = img.resize((width, height))
            else:
                width, height = pixels_exceed_limit(img.size[0], img.size[1])
                img = img.resize((width, height))
            img.save(buf, format=fmt)
            return buf.getvalue()

    elif _isinstance(data, "torch", "Tensor"):
        # pytorch tensor: convert to png
        from .. import requirePackage

        im = data

        with io.BytesIO() as buf:
            img = requirePackage("torchvision", "transforms").ToPILImage()(im)
            if width and height:
                rate = img.size[0] / img.size[1]
                ass_height = width / rate
                ass_width = rate * height
                if ass_height > height:
                    width = math.floor(height * rate)
                if ass_width > width:
                    height = math.floor(width / rate)
                img = img.resize((width, height))
            else:
                width, height = pixels_exceed_limit(img.size[0], img.size[1])
                img = img.resize((width, height))
            img.save(buf, format=fmt)
            return buf.getvalue()

    elif _isinstance(data, "tensorflow.python.framework.ops", "EagerTensor"):
        im = data
        return to_content_buf(im.numpy())

    elif _isinstance(data, "PIL.Image", "Image"):
        # PIL/Pillow images
        img = data

        with io.BytesIO() as buf:
            if width and height:
                rate = img.size[0] / img.size[1]
                ass_height = width / rate
                ass_width = rate * height
                if ass_height > height:
                    width = math.floor(height * rate)
                if ass_width > width:
                    height = math.floor(width / rate)
                img = img.resize((width, height))
            else:
                width, height = pixels_exceed_limit(img.size[0], img.size[1])
                img = img.resize((width, height))
            img.save(buf, format=fmt)
            return buf.getvalue()

    elif _isinstance(data, "matplotlib.figure", "Figure"):
        # matplotlib figures
        fig = data
        if fig.canvas is None:
            from matplotlib.backends.backend_agg import FigureCanvasAgg

            FigureCanvasAgg(fig)

        with io.BytesIO() as buf:
            fig.savefig(buf)
            if width and height:
                img = requirePackage("PIL", "Image", "Pillow").open(buf)
                rate = img.size[0] / img.size[1]
                ass_height = width / rate
                ass_width = rate * height
                if ass_height > height:
                    width = math.floor(height * rate)
                if ass_width > width:
                    height = math.floor(width / rate)
                img = img.resize((width, height))
            else:
                width, height = pixels_exceed_limit(img.size[0], img.size[1])
                img = img.resize((width, height))
            buf = io.BytesIO()
            img.save(buf, format=fmt)
            return buf.getvalue()

    else:
        raise TypeError("Unsupported type : {}".format(type(data)))


def imgcat(
    buf,
    width_scale=None,
    height_scale=None,
    preserve_aspect_ratio=True,
    fp=None,
):
    """
    Print image on terminal (iTerm2).

    Follows the file-transfer protocol of iTerm2 described at
    https://www.iterm2.com/documentation-images.html.

    Args:
        :param buf: the content of image in buffer interface, numpy array, etc.
        :param width: the width for displaying image, in number of characters (columns)
        :param height: the height for displaying image, in number of lines (rows)
        :param fp: The buffer to write to, defaults sys.stdout
        :param filename:
    """
    if fp is None:
        fp = sys.stdout.buffer  # for stdout, use buffer interface (py3)

    if len(buf) == 0:
        raise ValueError("Empty buffer")

    if is_iterm2:
        # now starts the iTerm2 file transfer protocol.
        fp.write(OSC)
        fp.write(b"1337;File=inline=1")
        fp.write(b";size=" + str(len(buf)).encode())
        if width_scale:
            fp.write(b";width=" + f'{width_scale}%%%%'.encode())
        if height_scale:
            fp.write(b";height=" + f'{height_scale}%%%%'.encode())
        if not preserve_aspect_ratio:
            fp.write(b";preserveAspectRatio=0")
        fp.write(b";inline=1")
        fp.write(b":")
        fp.flush()
        fp.write(base64.b64encode(buf))
        fp.write(ST)
        fp.write(b"\n")
        # flush is needed so that the cursor control sequence can take effect
        fp.flush()
    elif is_kitty:
        # Kitty graphics protocol
        # https://sw.kovidgoyal.net/kitty/graphics-protocol.html
        from base64 import standard_b64encode
        def serialize_gr_command(**cmd):
            payload = cmd.pop('payload', None)
            cmd = ','.join(f'{k}={v}' for k, v in cmd.items())
            ans = []
            w = ans.append
            w(b'\033_G'), w(cmd.encode('ascii'))
            if payload:
                w(b';')
                w(payload)
            w(b'\033\\')
            return b''.join(ans)

        def write_chunked(**cmd):
            data = standard_b64encode(cmd.pop('data'))
            while data:
                chunk, data = data[:4096], data[4096:]
                m = 1 if data else 0
                fp.write(serialize_gr_command(payload=chunk, m=m, **cmd))
                fp.flush()
                cmd.clear()
            fp.write(b"\n")
            fp.flush()
        write_chunked(a='T', f=100, data=buf)
    else:
        raise RuntimeError(
            "This function is only supported in iTerm2. "
        )

def image_preview(
    img,
    is_url=False,
    set_proxy: str = "",
    set_referer: str = "",
):
    """
    在终端预览图片 | 目前仅有MacOS下的iTerm可用, 但你可以开启强制显示选项预览

    preview image on terminal | At present, only iTerm under MacOS is available, but you can enable force show option to preview

    :param is_url: 是否为url
    :param img: opened file, numpy array, PIL.Image, matplotlib fig
    :param set_proxy: set proxy
    :param set_referer: set refer
    :return:
    """
    from .. import requirePackage

    try:
        bypass_flag = False
        if not is_url and (isinstance(img, str) and not os.path.exists(img)):
            is_url = img.startswith("http")

        if is_url:
            if img_bytes := requirePackage(".NetTools.NormalDL", "normal_dl")(
                img,
                set_referer=set_referer,
                set_proxy=set_proxy,
                disableStatus=True,
                write_to_memory=True
            ):
                if img.endswith('.svg'):
                    img_bytes = requirePackage("cairosvg", "svg2png")(bytestring=img_bytes)
                    bypass_flag = True
                img = img_bytes
            else:
                return qs_default_console.print(
                    requirePackage('.', 'qs_error_string'),
                    "Failed to download image" if user_lang != "zh" else "图片下载失败",
                )
        elif isinstance(img, str) and os.path.exists(img):
            if img.endswith('.svg') or img.endswith('.svgz'):
                qs_default_console.print(requirePackage('.', 'qs_warning_string'), 'Convert svg to png (dpi=300) ...' if user_lang != 'zh' else 'svg将转换为png (dpi=300) ...')
                img = requirePackage("io", "BytesIO")(requirePackage("cairosvg", "svg2png")(url=img, dpi=300))
                img = requirePackage("PIL", "Image", "Pillow").open(img)
            elif img.endswith('.eps') or img.endswith('.epsf') or img.endswith('.epsi'):
                qs_default_console.print(requirePackage('.', 'qs_warning_string'), 'Convert eps to png (dpi=300) ...' if user_lang != 'zh' else 'eps将转换为png (dpi=300) ...')
                img = requirePackage("io", "BytesIO")(requirePackage("wand.image", "Image")(filename=img).make_blob())
                img = requirePackage("PIL", "Image", "Pillow").open(img)
            elif img.endswith('.pdf'):
                qs_default_console.print(requirePackage('.', 'qs_warning_string'), 'Convert pdf to png ...' if user_lang != 'zh' else 'pdf将转换为png ...')
                with requirePackage('pdfplumber', 'open')(img) as pdf:
                    img = pdf.pages[0].to_image(resolution=300).original
            else:
                img = requirePackage("PIL", "Image", "Pillow").open(img)
        else:
            qs_default_console.print(
                requirePackage('.', 'qs_warning_string'),
                '将自适应解析图片格式' if user_lang != 'zh' else 'Will adaptively parse the image format',
            )

        _st = qs_default_status.status

        qs_default_status(
            "Loading image" if user_lang != "zh" else "正在加载图片",
        ).start()

        import array, fcntl, termios
        buf = array.array('H', [0, 0, 0, 0])
        fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, buf)
        theight, twidth, width, height = buf[0], buf[1], buf[2], buf[3]
        th_pixels = math.floor(height / theight)
        tw_pixels = math.floor(width / twidth)
        
        if tw_pixels and th_pixels:
            buf = to_content_buf(img, bypass=bypass_flag, width=width, height=height - (4 if is_kitty else 3) * th_pixels)
            iwidth, _ = get_image_shape(buf)
            tiwidth = math.ceil(iwidth / tw_pixels)
            pre_space = ' ' * ((twidth - tiwidth) // 2)
            qs_default_status.stop()
            print(pre_space, end='')
            sys.stdout.flush()
            imgcat(buf)
        else:
            # 缺乏终端尺寸信息，直接显示
            qs_default_status.stop()
            buf = to_content_buf(img, bypass=bypass_flag)
            imgcat(buf)

        if _st:
            qs_default_status.start()
    except Exception:
        qs_default_status.stop()
        qs_default_console.print_exception()
        return



# coding=utf-8
"""
qs的http服务

HTTP service of QS
"""
from socketserver import ThreadingMixIn
from http.server import SimpleHTTPRequestHandler
from http.server import HTTPServer
from .. import qs_default_console, qs_info_string, requirePackage


class HttpServers:
    qrcode_terminal = requirePackage("qrcode_terminal", real_name="qrcode-terminal")
    import signal

    def __init__(self, ip="localhost", port=8000, url=""):
        """
        http服务类初始化

        HTTP service class initialization

        :param ip: 绑定的ip
        :param port: 端口
        :param url: 对外显示的访问url
        """
        self.web_address = ip
        self.web_port = port
        self.httpd = None
        self.bind_url = url

    class Server(ThreadingMixIn, HTTPServer):
        pass

    def start(self):
        """
        开启http服务

        Start the HTTP service

        :return: None
        """
        HttpServers.signal.signal(HttpServers.signal.SIGINT, self.shutdown)
        self.httpd = HttpServers.Server(
            (self.web_address, self.web_port), SimpleHTTPRequestHandler
        )
        if not self.bind_url:
            self.bind_url = "http://" + self.web_address + ":" + str(self.web_port)
        qs_default_console.print(qs_info_string, self.bind_url)  # * 展示待访问的url
        HttpServers.qrcode_terminal.draw(self.bind_url)  # * 为待访问的url绘制二维码
        try:
            self.httpd.serve_forever()
        except TypeError:
            self.shutdown()

    def shutdown(self):
        """
        Ctrl C
        :return: None
        """
        import os

        self.httpd.shutdown()
        qs_default_console.print(qs_info_string, "HTTP Server: Closed.")
        os._exit(0)



from . import size_format, get_fileinfo, headers
from concurrent.futures import ThreadPoolExecutor, wait
from ..SystemTools import get_core_num
from ..TuiTools.Bar import NormalProgressBar
from .. import (
    user_lang,
    qs_default_console,
    qs_default_status,
    qs_error_string,
    qs_info_string,
    qs_warning_string,
)
from threading import Lock
from requests import get
import queue
import time
import os

core_num = min(get_core_num() * 4, 16)


class MultiSingleDL:
    def __init__(
        self,
        urls: list,
        rt_dir: str,
        proxy: str = "",
        referer: str = "",
        save_to_mem: bool = False,
    ):
        self.proxy = proxy
        self.referer = referer
        self.proxies = (
            {"http": "http://" + proxy, "https": "https://" + proxy} if proxy else {}
        )
        self.headers = headers
        if referer:
            self.headers["Referer"] = referer
        self.rt_dir = rt_dir
        self.save_to_mem = save_to_mem
        self.status_dict = {}
        self.infos = {}
        self.urls = urls.copy()
        self.task_num = len(self.urls)
        if self.save_to_mem:
            self.content_ls = [b""] * self.task_num
        self.cur_task_num = 0
        self.task_num_lock = Lock()
        self.max_retry = 3
        self.job_queue = queue.Queue()
        self.pool = None
        self.status = qs_default_status("获取文件信息中...")
        self.progress, self.task_id = None, None

    def _info(self, url):
        retry = 3
        while retry:
            real_url, name, r = get_fileinfo(url, self.proxy, self.referer)
            if all([real_url, name, r]):
                break
            retry -= 1

        info_flag = True
        if not (real_url and name and r):
            info_flag = False
            qs_default_console.print(
                qs_error_string if self.failed2exit else qs_warning_string,
                f"{url}: Get File information failed, please check network!"
                if user_lang != "zh"
                else "获取文件信息失败，请检查网络!",
                (real_url, name, r)
            )

        self.infos[url] = {
            "url": real_url,
            "name": os.path.basename(url) if name and "." not in name else name,
            "size": -1
            if not info_flag
            else int(r.headers["content-length"])
            if "content-length" in r.headers
            else -1,
        }

        self.task_num_lock.acquire()
        self.cur_task_num += 1
        total = sum([self.infos[i]["size"] for i in self.infos if self.infos[i]["size"] != -1])
        self.status.update(f"获取文件信息中 [bold cyan]{self.cur_task_num}[/]/[bold]{self.task_num}[/]: [bold green]{size_format(total)}[/]")
        self.task_num_lock.release()

    def _dl(self, url, filename, job_id, retry=0):
        try:
            r = get(
                self.infos[url]["url"],
                stream=True,
                proxies=self.proxies,
                headers=self.headers,
            )
            if not self.save_to_mem:
                with open(os.path.join(self.rt_dir, filename), "wb") as f:
                    for chunk in r.iter_content(32768):
                        f.write(chunk)
                        self.progress.advance(self.task_id, 32768)
            else:
                for chunk in r.iter_content(32768):
                    self.content_ls[job_id] += chunk
                    self.progress.advance(self.task_id, 32768)
        except Exception as e:
            if retry < self.max_retry:
                self.progress.print(
                    qs_warning_string,
                    f'"{url}": task anomaly, ' if user_lang != "zh" else f'"{url}": 任务异常, ',
                    "retrying..." if user_lang != "zh" else "正在重试...",
                )
                self.job_queue.put(
                    {
                        "url": url,
                        "filename": filename,
                        "job_id": job_id,
                        "retry": retry + 1,
                    }
                )
            else:
                self.progress.print(
                    qs_error_string,
                    f'"{url}":',
                    "Download failed!" if user_lang != "zh" else "下载失败！",
                )
                self.status_dict[url] = "failed"
                self.progress.advance(self.task_id, 1)

    def run(self, name_map: dict = None, qps: int = None, qps_info:int = None, qps_download: int = None, without_output: bool = False, without_info: bool = False):
        """
        :param name_map: {url: filename}
        :param qps: qps for download
        :param qps_info: qps for get file info
        :param qps_download: qps for download
        :param without_output: without output
        :param without_info: without get file info
        """
        self.qps = qps
        qps_info = qps_info if qps_info else qps
        qps_download = qps_download if qps_download else qps

        self.status.start()
        if qps_info:
            self.pool = ThreadPoolExecutor(max_workers=qps_info)
            wait([self.pool.submit(self._info, item) for item in self.urls])
        else:
            self.pool = ThreadPoolExecutor(max_workers=core_num)
            wait([self.pool.submit(self._info, item) for item in self.urls])
        self.pool.shutdown(wait=True)

        if name_map:
            from .. import requirePackage
            file_suffix = requirePackage('.SystemTools', 'file_suffix')

        total = sum([self.infos[i]["size"] for i in self.infos])
        failed_num = [self.infos[i]["size"] for i in self.infos].count(-1)
        if not without_output:
            qs_default_console.print(
                qs_info_string,
                f"获取文件信息完毕! 文件总大小: {size_format(total)} | 未获取到详细信息条数为: {failed_num}",
            )
        self.status.stop()
        
        self.progress, self.task_id = NormalProgressBar("多文件下载", total)

        for _id, item in enumerate(self.urls):
            self.job_queue.put(
                {"url": item,
                    "filename": self.infos[item]["name"], "job_id": _id}
                if not name_map
                else {"url": item, "filename": f'{name_map[item]}.{file_suffix(self.infos[item]["name"])}', "job_id": _id}
            )

        self.progress.start()
        wait_list = []
        self.pool = ThreadPoolExecutor(max_workers=qps_download) if qps_download else ThreadPoolExecutor(max_workers=core_num)
        while not self.job_queue.empty():
            while not self.job_queue.empty():
                job = self.job_queue.get()
                wait_list.append(self.pool.submit(self._dl, **job))
            wait(wait_list)
        self.progress.stop()
        return (
            [os.path.join(self.rt_dir, self.infos[i]["name"])
             for i in self.infos]
            if not name_map
            else [os.path.join(self.rt_dir, f'{name_map[i]}.{file_suffix(self.infos[i]["name"])}') for i in self.infos]
        )

    def get_content_ls(self):
        return self.content_ls


def multi_single_dl(
    urls: list,
    rt_dir: str = "./",
    proxy: str = "",
    referer: str = "",
    name_map: dict = None,
    qps: int = 0,
    qps_info: int = 0,
    qps_download: int = 0,
    without_output: bool = False,
):
    """
    并行多个小文件下载

    :param urls: 小文件的URL
    :param rt_dir: 文件下载目录
    :param proxy: 代理
    :param referer: referer
    :param name_map: 文件命名映射{URL: filename}, 后缀名会自动获取并添加
    :param qps: 限制每秒请求次数, qps_info和qps_download默认为qps
    :param qps_info: 限制每秒获取信息次数
    :param qps_download: 限制每秒下载次数
    :param without_output: 不输出信息
    :return: 下载好的文件路径列表
    """
    return MultiSingleDL(
        urls=urls, rt_dir=rt_dir, proxy=proxy, referer=referer
    ).run(name_map=name_map, qps=qps, qps_info=qps_info, qps_download=qps_download, without_output=without_output)


def multi_single_dl_content_ls(
    urls: list,
    rt_dir: str = "./",
    proxy: str = "",
    referer: str = "",
    qps: int = 0,
    qps_info: int = 0,
    qps_download: int = 0,
    without_output: bool = False,
):
    """
    并行多个小文件下载

    :param urls: 小文件的URL
    :param rt_dir: 文件下载目录
    :param proxy: 代理
    :param referer: referer
    :param qps: 限制每秒请求次数
    :param qps_info: 限制每秒获取信息次数
    :param qps_download: 限制每秒下载次数
    :param without_output: 不输出信息
    :return: 存储在内存中的文件内容列表
    """
    dl = MultiSingleDL(
        urls=urls,
        rt_dir=rt_dir,
        proxy=proxy,
        referer=referer,
        save_to_mem=True,
    )
    dl.run(qps=qps, qps_info=qps_info, qps_download=qps_download, without_output=without_output)
    return dl.get_content_ls()



# coding=utf-8
"""
Docs for different language:
    中文: https://rhythmlian.cn/2020/08/09/QuickStart-Rhy-zh/

    English: https://rhythmlian.cn/2020/02/14/QuickStart-Rhy/
"""
import os
import sys
import time

from QuickProject import QproDefaultStatus as qs_default_status
from QuickProject import external_exec, user_pip, user_root
from QuickProject import QproInfoString as qs_info_string
from QuickProject import QproWarnString as qs_warning_string
from QuickProject import QproErrorString as qs_error_string


from .__cache__ import QsCache
from .__config__ import (
    QsConfig,
    dir_char,
    platform,
    qs_default_console,
    _ask,
    user_lang,
    lang_detector
)

name = "QuickStart_Rhy"
qs_config = QsConfig(os.path.join(user_root, ".qsrc"))
qs_cache = QsCache(os.path.join(user_root, ".qs_cache"))

trans_engine = qs_config.basicSelect("default_translate_engine")["support"][
    qs_config.basicSelect("default_translate_engine")["index"]
]
qs_console_width = qs_default_console.width

_package_info_ = qs_cache.get("package_info")
_not_update_ = ["io", "tarfile", "zipfile", "shutil"]
if _package_info_ is None:
    _package_info_ = {}

def requirePackage(
    pname: str,
    module: str = "",
    real_name: str = "",
    not_ask: bool = False,
    set_pip: str = user_pip,
    keep_latest: bool = False,
):
    """
    获取本机上的python第三方库

    :param pname: 库名
    :param module: 待引入的模块名，可缺省
    :param real_name: 用于 pip3 install 的名字
    :param not_ask: 不询问
    :param set_pip: pip3的路径
    :param keep_latest: 是否保持最新版本
    :return: 库或模块的地址
    """
    local_scope = {}
    try:
        package_name = pname.split(".")[0] if not real_name else real_name
        if not package_name:  # 引用为自身
            package_name = name
        if package_name not in _not_update_ and (
            keep_latest
            and time.time() - _package_info_.get(package_name, 0) > 3600 * 24 * 7
        ):
            with qs_default_status(f"{lang_detector['updating']} {package_name}"):
                _st, _ct = external_exec(
                    f"{set_pip} install {package_name} -U", without_output=True
                )
                if _st:
                    qs_default_console.print(
                        qs_warning_string,
                        f"{lang_detector['updating']} {package_name} {lang_detector['fail_and_execute_manually']}"
                        f"'{set_pip} install {package_name} -U'",
                    )
                    qs_default_console.print(qs_warning_string, _ct)
            _package_info_[package_name] = time.time()
            qs_cache.set("package_info", _package_info_)
        
        exec((f"from {pname} import {module}" if module else f"import {pname}"), globals(), local_scope)
    except (ModuleNotFoundError, ImportError):
        if not_ask:
            return None
        if package_name == name:
            qs_default_console.print(qs_warning_string, lang_detector['requirePackageWarning'])
        if _ask(
            {
                "type": "confirm",
                "message": f"""Qs {lang_detector['require']} {pname + (' -> ' + module if module else '')}, {lang_detector['confirm_install']}?""",
                "default": True,
            }
        ):
            with qs_default_status(
                f"{lang_detector['Install']} {package_name}"
            ):
                st, _ = external_exec(
                    f"{set_pip} install {package_name} -U",
                    True,
                )
            if st:
                qs_default_console.print(
                    qs_error_string,
                    f"{lang_detector['Install']} {pname + (' -> ' + module if module else '')} {lang_detector['fail_and_execute_manually']}",
                    f"'{set_pip} install {package_name} -U'",
                )
                exit(-1)
            exec((f"from {pname} import {module}" if module else f"import {pname}"), globals(), local_scope)
    return local_scope.get(module if module else pname)


def cut_string(string: str, length: int, ignore_charset: list = []) -> list:
    """
    每隔l个字符切分字符串

    :param string: 字符串
    :param length: 切分长度
    :param ignore_charset: 忽略的字符集
    :return: 切分后产生的list
    """
    res, cur, cnt = [], "", 0
    for i in string:
        _char_len = 0 if i in ignore_charset else 2 if ord(i) > 255 else 1
        cnt += _char_len
        if cnt <= length:
            cur += i
        else:
            res.append(cur)
            cur, cnt = i, _char_len
    if cur:
        res.append(cur)
    return res


def _is_in_path(cmd: str):
    """
    判断cmd是否在环境变量中

    :param cmd: 命令
    :return: 是否在环境变量中
    """
    import os

    ls = os.environ["PATH"].split(os.pathsep)
    for prefix in ls:
        if os.path.exists(os.path.join(prefix, cmd)):
            return True
    return False


def table_cell(string: str, length: int) -> str:
    return "\n".join(cut_string(string, length))


def deal_ctrl_c(signum, frame):
    """
    默认的处理ctrl c的函数

    default function to deal [CTRL C]

    :param signum: signum
    :param frame: frame
    :return: None
    """
    if signum or frame or True:
        exit(0)


def remove(path):
    """
    删除文件或文件夹

    delete file or folder.

    :param path: 路径
    :return: None
    """
    if os.path.exists(path):
        if os.path.isdir(path):
            import shutil

            shutil.rmtree(path)
        else:
            os.remove(path)


def cur_time():
    """
    获取当前时间

    get time of now

    :return: None
    """
    tm = time.strftime("%Y-%m-%d %A %H:%M:%S", time.localtime(time.time())).split()
    qs_default_console.print(qs_info_string, " ".join(tm))


def open_url(argv: list = None):
    """
    打开命令行参数中的链接

    open urls in argv

    :return: None
    """
    import webbrowser as wb
    from .NetTools import formatUrl

    if not argv:
        argv = sys.argv[2:]
    if argv:
        for url in argv:
            url = formatUrl(url)
            wb.open_new_tab(url)
    else:
        pyperclip = requirePackage("pyperclip")
        try:
            url = pyperclip.paste()
        except:
            url = _ask(
                {
                    "type": "input",
                    "message": lang_detector['pyperclip_not_support']
                }
            )
        wb.open_new_tab(formatUrl(url))


def open_app():
    """
    Mac OS下的open（不支持其他平台）

    open app (only support Mac OS)

    :return: None
    """
    if platform == "darwin":
        external_exec('open -a "' + '" "'.join(sys.argv[2:]) + '"')
    else:
        return qs_default_console.print(
            qs_error_string,
            lang_detector['macos_only']
        )


def open_file(*argv):
    """
    使用合适应用打开文件

    Open file with appropriate application

    :return: None
    """
    if not argv:
        argv = sys.argv[2:]
    if platform == "darwin":
        external_exec('open "' + '" "'.join(argv) + '"')
    elif platform == "linux":
        external_exec('xdg-open "' + '" "'.join(argv) + '"')
    else:
        for file in argv:
            os.startfile(os.path.abspath(file))


def calculate():
    """
    执行算数表达式

    Execute the arithmetic expression.

    :return: None
    """
    try:
        exp = " ".join(sys.argv[2:])
        qs_default_console.print("%s = %s" % (exp, eval(exp)))
    except Exception as e:
        qs_default_console.print(qs_error_string, repr(e))


def pcat():
    """
    输出粘贴板中内容

    Output the contents of the clipboard
    :return: None
    """
    print(requirePackage("pyperclip", "paste")())


def fcopy():
    """
    获取文件内容并复制进粘贴板

    Get the content of the file and copy it into the pasteboard
    :return:
    """
    if not os.path.exists(sys.argv[2]):
        return qs_default_console.print(
            qs_error_string,
            "No such file:" if user_lang != "zh" else "未找到文件:",
            sys.argv[2],
        )
    with open(sys.argv[2], "r") as f:
        requirePackage("pyperclip").copy(f.read())


def copy():
    """
    复制文件到粘贴板
    :return:
    """

    def which(command):
        return False if external_exec("which %s" % command, True)[0] else True

    if not os.path.exists(sys.argv[2]):
        return qs_default_console.print(
            qs_error_string,
            "No such file:" if user_lang != "zh" else "未找到文件:",
            sys.argv[2],
        )
    if platform == "darwin":
        # 检查 pbadd 是否在 PATH 中
        if not which("pbadd"):
            from QuickStart_Rhy.NetTools.NormalDL import normal_dl

            normal_dl(
                "https://cos.lhc.studio/ImgBed/86438ea0f489a2c75ff7263eda630005",
                set_name="pbadd",
            )
            external_exec("chmod +x pbadd")
            external_exec("mv pbadd /usr/local/bin/")
        external_exec("pbadd " + sys.argv[2], True)
    elif platform.startswith("win"):
        # 检查是否为 powershell
        st, _ = external_exec(
            'powershell -Command "echo $PSVersionTable.PSVersion"', True
        )
        if st:
            return qs_default_console.print(
                qs_error_string,
                "Please use powershell to run this command"
                if user_lang != "zh"
                else "请使用 powershell 运行此命令",
            )
        external_exec(
            'powershell -Command "Set-Clipboard -Path ' + sys.argv[2] + '"', True
        )
    else:
        return qs_default_console.print(
            qs_error_string,
            lang_detector['not_support_current_system']
        )


def tcopy():
    """
    复制文本到粘贴板
    :return:
    """
    requirePackage("pyperclip", "copy")(" ".join(sys.argv[2:]))


def get_user_lang():
    print(user_lang)


def play_music(*argv, using_afplay: bool = False):
    if not argv:
        argv = sys.argv[2:]

    if not using_afplay:
        AS = requirePackage("pydub", "AudioSegment")
        play = requirePackage("pydub.playback", "play")
    
    for music in argv:
        try:
            play(AS.from_file(music)) if not using_afplay else external_exec(
                f"afplay {os.path.abspath(music)}"
            )
        except:
            pass


def qs_print(*argv):
    """
    打印文件

    :return:
    """
    if not argv:
        argv = sys.argv[2:]
    if platform.startswith("win"):
        win32api = requirePackage("pywin32", "win32api")
        win32print = requirePackage("pywin32", "win32print")
        for file in argv:
            win32api.ShellExecute(
                0, "print", file, '/d:"%s"' % win32print.GetDefaultPrinter(), ".", 0
            )
    else:
        for file in argv:
            external_exec(f"lp {file}")


def sas():
    """
    设置音源

    Set audio source

    :return: None
    """
    if platform != "darwin":
        qs_default_console.print(
            qs_error_string,
            "Not support your system" if user_lang != "zh" else "不支持你的系统",
        )
        return
    if not _is_in_path("SwitchAudioSource"):
        if _ask(
            {
                "type": "confirm",
                "message": "Install SwitchAudioSource?"
                if user_lang != "zh"
                else "安装 SwitchAudioSource？",
                "default": True,
            }
        ):
            with qs_default_status(
                "Installing SwitchAudioSource..."
                if user_lang != "zh"
                else "正在安装 SwitchAudioSource..."
            ):
                retries = 3
                while retries:
                    external_exec("brew install SwitchAudioSource")
                    if _is_in_path("SwitchAudioSource"):
                        break
                    retries -= 1
                if not retries:
                    qs_default_console.print(
                        qs_error_string,
                        "Install failed" if user_lang != "zh" else "安装失败",
                    )
                    return
    import json

    current = json.loads(external_exec("SwitchAudioSource -c -f json", True)[1])
    devices = [
        json.loads(i)
        for i in external_exec("SwitchAudioSource -a -f json", True)[1].splitlines()
    ]
    devices = {
        i["id"]: {
            "uid": i["uid"],
            "type": "🔈" if i["type"] == "output" else "🎤",
            "name": i["name"],
        }
        for i in devices
    }
    if not devices:
        qs_default_console.print(
            qs_error_string, "No device found" if user_lang != "zh" else "未找到设备"
        )
        return
    from .TuiTools.Table import qs_default_table

    table = qs_default_table(
        ["ID", "Name", "Type", "Current"]
        if user_lang != "zh"
        else ["ID", "名称", "类型", "当前设备"]
    )
    for i in devices:
        table.add_row(i, devices[i]["name"], devices[i]["type"]) if devices[i][
            "uid"
        ] != current["uid"] else table.add_row(
            i, devices[i]["name"], devices[i]["type"], "[bold green]⬅[/]"
        )
    qs_default_console.print(table, justify="center")

    default = qs_cache.get("audio_source")
    uids = {val["uid"]: key for key, val in devices.items()}
    question = {
        "type": "input",
        "message": "Input device ID" if user_lang != "zh" else "输入设备 ID",
        "validate": lambda x: x in [str(i) for i in devices],
    }
    if default:
        _default = default.copy()
        for item in default:
            if item not in uids:
                _default.pop(item)
        if _default:
            if len(_default) > 1:
                _default.pop(current["uid"])
            question["default"] = uids[max(_default, key=_default.get)]
    else:
        default = {}
    if not (select := _ask(question)):
        return
    uid = devices[select]["uid"]
    default[uid] = default.get(uid, 0) + 1

    external_exec("SwitchAudioSource -i %s" % select, True)
    qs_cache.set("audio_source", default)
    qs_default_console.print(qs_info_string, "Done" if user_lang != "zh" else "完成")


def swap(file1: str = None, file2: str = None):
    """
    交换两个文件的内容
    """
    index = 2
    if not file1 or not file2:
        file1 = sys.argv[index]
        index += 1
    if not file2:
        file2 = sys.argv[index]

    with open(file1, "rb") as f1:
        data1 = f1.read()
    with open(file2, "rb") as f2:
        data2 = f2.read()

    with open(file1, "wb") as f1:
        f1.write(data2)
    with open(file2, "wb") as f2:
        f2.write(data1)

requirePackage("QuickProject", real_name="Qpro") # 自动更新

def wrap_text_preserve_links(text, width, with_numba=False, with_raw=False):
        import re
        if with_numba:
            from .NumbaTools import cut_string
        else:
            from . import cut_string
        
        def split_line(line):
            parts = []
            current_part = ""
            link_pattern = r'(!?\[([^\]]+)\]\(([^\)]+)\))'
            
            for match in re.finditer(link_pattern, line):
                before_link = line[len(current_part):match.start()]
                parts.extend(cut_string(before_link, width, ignore_charset="`"))
                parts.append('\n\n' + match.group(0) + '\n\n')
                current_part = line[:match.end()]
            
            remaining = line[len(current_part):]
            parts.extend(cut_string(remaining, width, ignore_charset="`"))
            return ' '.join(parts)

        wrapped_lines = []
        for line in text.split('\n'):
            wrapped_lines.append(split_line(line))
        
        if with_raw:
            return wrapped_lines
        return '\n'.join(wrapped_lines)

def live_show(content, is_stream: bool = True, prefix: str = ""):
    if is_stream:
        from rich.live import Live
        from rich.markdown import Markdown

        with Live(
            "",
            console=qs_default_console,
            auto_refresh=False,
            vertical_overflow="visible",
        ) as live:
            total_res = ""
            for res in content:
                res = prefix + res
                display = wrap_text_preserve_links(res, qs_default_console.width, False)
                live.update(Markdown(display, justify="full"), refresh=True)
                total_res = res
    else:
        qs_default_console.print(wrap_text_preserve_links(content, qs_default_console.width, False))
        total_res = content
    return total_res



# coding=utf-8
"""
百度云相关API

Baidu cloud API
"""
import os
from . import pre_check, user_lang
from .. import qs_default_console, qs_error_string, qs_info_string, requirePackage

imgp = requirePackage("aip", "imageprocess", real_name="baidu-aip")
aip = requirePackage("aip", real_name="baidu-aip")


class ImageDeal:
    """
    百度图像处理类

    Baidu image processing class
    """

    import base64

    def __init__(
        self,
        app_id: str = pre_check("AipImageAPP_ID"),
        app_key: str = pre_check("AipImageAPP_KEY"),
        secret_key: str = pre_check("AipImageSECRET_KEY"),
    ):
        """
        初始化并登陆百度图像处理

        Initialize and log in Baidu image processing

        :param app_id: 应用ID
        :param app_key: 应用key
        :param secret_key: 密钥
        """
        self.client = imgp.AipImageProcess(app_id, app_key, secret_key)

    def largeImage(self, path: str):
        """
        放大图片 (图像效果增强)

        Enlarge image (Image Enhancement)

        :param st: console状态
        :param path: 图片路径
        :return: None
        """
        if not os.path.exists(path) or not os.path.isfile(path):
            qs_default_console.log(
                qs_error_string,
                ("No file named: %s" if user_lang != "zh" else "没有文件: %s") % path,
            )
            return
        img_name = os.path.basename(path)
        img_name = (
            img_name[: img_name.index(".")] + "_LG." + ".".join(img_name.split(".")[1:])
        )

        from .. import qs_default_status

        with qs_default_status(
            f'{"Reading Image" if user_lang != "zh" else "读取图片"}: {path}'
        ) as status:
            with open(path, "rb") as f:
                img = f.read()
            status.update(
                f'{qs_info_string} {"Dealing..." if user_lang != "zh" else "处理中..."}'
            )
            img = self.client.imageQualityEnhance(img)
            try:
                status.update(
                    f'{qs_info_string} {"Write to" if user_lang != "zh" else "写入"}: {img_name}'
                )
                img = ImageDeal.base64.b64decode(img["image"])
                with open(img_name, "wb") as f:
                    f.write(img)
            except:
                qs_default_console.print(qs_error_string, img)
        qs_default_console.print(
            qs_info_string, "Deal Done!" if user_lang != "zh" else "处理完成!"
        )


class AipNLP:
    """
    百度语言处理类

    Baidu language processing class
    """

    def __init__(
        self,
        appid: str = pre_check("AipNlpAPP_ID"),
        appkey: str = pre_check("AipNlpAPP_KEY"),
        sckey: str = pre_check("AipNlpSECRET_KEY"),
    ):
        """
        初始化并登陆百度语言处理应用

        Initialize and log in baidu language processing application

        :param appid: 应用ID
        :param appkey: 应用Key
        :param sckey: 密钥
        """
        self.client = aip.AipNlp(appid, appkey, sckey)

    def get_res(self, words: str):
        """
        文本纠错

        Text error correction

        :param words: 待处理文本
        :return: 处理后的文本 | proceed words
        """
        try:
            res = self.client.ecnet(words)["item"]["correct_query"]
        except Exception as e:
            qs_default_console.log(qs_error_string, repr(e))
        else:
            return res



# coding=utf-8
from . import *
from .. import user_currency
import json
from urllib.parse import quote
import requests


alapi_token = pre_check("alapi_token", ext=False)
if not alapi_token:
    from .. import qs_default_console, qs_error_string, qs_warning_string, _ask

    qs_default_console.print(
        qs_error_string, "This function require alapi token!\n这个功能需要alapi token\n"
    )
    if _ask(
        {
            "type": "confirm",
            "message": "To get alapi token?\n是否前往获取alapi token?",
            "default": True,
        }
    ):
        from .. import open_url
        from .. import qs_config

        try:
            open_url(["https://www.alapi.cn/"])
        except Exception as e:
            qs_default_console.print(
                qs_error_string,
                "Failed to open browser, try url: https://www.alapi.cn/",
            )
        finally:
            alapi_token = _ask(
                {
                    "type": "input",
                    "message": "Input alapi token | 输入alapi token:",
                }
            )
            qs_config.apiUpdate("alapi_token", alapi_token)
    else:
        exit()


v2_url = "https://v3.alapi.cn/api/"


def translate(text: str, from_lang: str = "auto", target_lang: str = user_lang):
    """
    获取翻译结果

    Get the translation results.

    :param text: 待翻译内容 | Content to be translated.
    :param from_lang: 语种来源 | Source language
    :param target_lang: 翻译成的语种 | Translated into the language
    :return: 翻译的文本 | Translated text
    """
    global alapi_token
    try:
        request_info = (
            "q={}&from={}&to={}".format(quote(text, "utf-8"), from_lang, target_lang)
            if from_lang
            else "q={}&to={}".format(quote(text, "utf-8"), target_lang)
        )
        res = requests.post(
            v2_url + "fanyi",
            data=request_info,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            }
            if alapi_token
            else {"Content-Type": "application/x-www-form-urlencoded"},
        )
        if res.status_code == requests.codes.ok:
            res = json.loads(res.text)
            if res["code"] != 200:
                return "[ERROR] {}".format(res["msg"])
            return res["data"]["dst"]
        return "[ERROR] 未知错误 | Unknown Error"
    except Exception as unknown_error:
        return f"[ERROR] {repr(unknown_error)}"


def bili_cover(url: str):
    """
    获取BiliBili视频封面

    Get the BiliBili video cover

    :param url: BiliBili视频链接或视频号 | BiliBili video link or video number
    :return:
    """
    import re
    from ..NetTools.NormalDL import normal_dl
    from .. import qs_default_console, qs_error_string, qs_info_string
    from ..ImageTools.ImagePreview import image_preview

    try:
        res = requests.post(
            v2_url + "bilibili/cover",
            data="c=" + url,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            },
        )
    except Exception as e:
        return qs_default_console.print(qs_error_string, repr(e))
    if res.status_code == requests.codes.ok:
        res = json.loads(res.text)
        if res["code"] != 200 or res["msg"] != "success":
            qs_default_console.log(
                qs_error_string,
                ("Get cover with: %s failed:" if user_lang != "zh" else "下载封面: %s 失败: ")
                % url,
            )
            return
        res = res["data"]
        res["description"] = res["description"].replace("<br />", "\n\t")
        res["description"] = res["description"].replace("&nbsp;", " ")
        res["description"] = re.sub("<.*?>", "", res["description"]).strip()  # 忽略HTML标签
        qs_default_console.print(
            qs_info_string,
            "TITLE:" if user_lang != "zh" else "标题:",
            "%s" % res["title"],
        )
        qs_default_console.print(
            qs_info_string, "" if user_lang != "zh" else "简介:", end="\n\t"
        )
        qs_default_console.print(res["description"], end="\n\n")
        normal_dl(res["cover"], res["title"] + "." + res["cover"].split(".")[-1])
        image_preview(res["title"] + "." + res["cover"].split(".")[-1])
    else:
        qs_default_console.log(
            qs_error_string,
            f"Get cover with: {url} failed" if user_lang != "zh" else f"下载封面: {url} 失败",
        )


def ip_info(ip: str):
    """
    获取ip的运营商、地理位置等数据

    Get IP operator, geographic location and other data

    :param ip: ipv4, ipv6, empty means current machine
    :return: data dict {ip, isp, pos | ERROR MESSAGE, location | ERROR code}
    """
    try:
        res = requests.get(
            v2_url + ("ip?ip=%s&format=json" % ip if ip else "ip"),
            headers={"token": alapi_token},
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return {
                "ip": ip,
                "isp": "",
                "pos": res["msg"],
                "location": "{code, %s}" % res["code"],
            }
        return res["data"]
    except Exception as e:
        return {
            "ip": ip,
            "isp": "",
            "pos": "Network Error" if user_lang != "zh" else "网络错误",
            "location": "{code, %s}" % repr(e),
        }


def garbage_classification(query_ls: list):
    """
    查询中国垃圾分类

    Search Chinese garbage classification

    :param query_ls: 待查询的垃圾列表 | garbage list
    :return: 查询结果的字符串表格 | string table
    """
    from .. import qs_default_console, cut_string

    width = qs_default_console.width // 5 * 2 - 1

    def fmt_string(string, pre_num):
        return "\n" * pre_num + " ".join(cut_string(string, width))

    from ..TuiTools.Table import qs_default_table
    from rich.text import Text
    import math

    table = qs_default_table(
        [
            {"header": "名称", "justify": "center", "style": "bold cyan"},
            {"header": "分类", "justify": "center"},
            {"header": "解释", "justify": "center"},
            {"header": "提示", "justify": "center"},
        ],
        "[bold underline] 查询结果\n",
    )
    first_flag = True
    for query_el in query_ls:
        if first_flag:
            first_flag = False
        else:
            table.add_row(Text("-", style="none"), Text("-", style="none"), "-", "-")
        res = requests.post(
            v2_url + "garbage",
            data="name={}".format(quote(query_el, "utf-8")),
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            }
            if alapi_token
            else {"Content-Type": "application/x-www-form-urlencoded"},
        )
        if res.status_code == requests.codes.ok:
            res = json.loads(res.text)
            if res["code"] != 200:
                table.add_row(
                    fmt_string(query_el, 0), "", "", fmt_string(res["msg"], 0)
                )
                continue
            try:
                res = res["data"][0]
                half_rows = [
                    math.ceil(len(query_el) / width / 2),
                    1,
                    math.ceil(len(res["explain"]) / width / 2),
                    math.ceil(len(res["tip"]) / width / 2),
                ]
                max_row = max(half_rows)
                half_rows = [max_row - i for i in half_rows]
                table.add_row(
                    fmt_string(query_el, half_rows[0]),
                    fmt_string(
                        [
                            "[bold green]可回收",
                            "[bold red]有害",
                            "[bold yellow]厨余(湿)",
                            "[bold blue]其他(干)",
                        ][int(res["type"]) - 1],
                        half_rows[1],
                    ),
                    Text(fmt_string(res["explain"], half_rows[2]), justify="full"),
                    Text(fmt_string(res["tip"], half_rows[3]), justify="full"),
                )
            except IndexError:
                table.add_row(fmt_string(query_el, 0), "Unknown", "未知垃圾", "未知垃圾")
        else:
            table.add_row(fmt_string(query_el, 0), "Unknown", "None", "Request Error")
    return table


def short_video_info(url: str):
    """
    解析多平台短视频，并返回视频信息与下载直链

    Parse multi - platform short video and return video information with download straight link

    :param url: 短视频分享链接 | Short video sharing link
    :return:
    """
    try:
        res = requests.post(
            v2_url + "video/url",
            data="url={}".format(quote(url, "utf-8")),
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            }
            if alapi_token
            else {"Content-Type": "application/x-www-form-urlencoded"},
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, {
                "title": res["msg"],
                "cover_url": "None",
                "video_url": "None",
                "source": str(res["code"]),
            }
        return True, res["data"]
    except Exception as e:
        return False, {
            "title": "网络错误" if user_lang == "zh" else "Network Error",
            "cover_url": "None",
            "video_url": "None",
            "source": "Unknown",
        }


def acg():
    """
    随机获取一张acg图片链接

    Get a random link to an ACG image

    :return: acg image link
    """
    try:
        res = requests.post(
            v2_url + "acg",
            data="format=json",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            }
            if alapi_token
            else {"Content-Type": "application/x-www-form-urlencoded"},
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"], 0, 0
        return True, res["data"]["url"], res["data"]["width"], res["data"]["height"]
    except Exception as e:
        return False, repr(e), 0, 0


def bingImg():
    """
    随机获取一张bing图片链接

    Get a link to a bing picture at random

    :return: image link
    """
    try:
        res = requests.post(
            v2_url + "bing",
            data="format=json",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            }
            if alapi_token
            else {"Content-Type": "application/x-www-form-urlencoded"},
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"], ""
        return True, res["data"]["url"], res["data"]["copyright"]
    except Exception as e:
        return False, repr(e), ""


def kdCheck(kd_number: str):
    """
    查询中国快递

    Inquire about China Express

    :param kd_number: 快递单号 | courier number
    :return: bool, list
    """
    try:
        res = requests.post(
            v2_url + "kd",
            data="number=%s" % kd_number,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            },
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, 0, res["msg"]
        return (
            True,
            int(res["data"]["state"]),
            res["data"]["info"],
        )
    except Exception as e:
        return False, 0, repr(e)


def pinyin(zh_str: str):
    """
    查询中文的拼音

    Query Chinese pinyin

    :param zh_str: Chinese | 中文
    :return:
    """
    try:
        res = requests.post(
            v2_url + "pinyin",
            data=f'word={quote(zh_str, "utf-8")}&tone=1&abbr=0',
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            },
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"]
        return True, res["data"]["pinyin"]
    except Exception as e:
        return False, repr(e)


def exchange(fr, number, to=user_currency):
    """
    查询<number> <fr> 对应 多少 <to>

    Query <number> <fr> corresponding <number?> <to>

    :param to: 待查询的币种
    :param number: 待查询的货币数量
    :param fr: 目标币种
    :return:
    """

    try:
        res = requests.get(
            v2_url + "exchange",
            params={"money": number, "from": fr, "to": to},
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            },
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"]
        return True, res["data"]
    except Exception as e:
        return False, repr(e)


def zhihuDaily():
    """
    获取知乎日报

    Get Zhihu Daily

    :return:
    """
    try:
        res = requests.get(
            v2_url + "zhihu",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            },
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"]
        return True, res["data"]
    except Exception as e:
        return False, repr(e)


def daily60s():
    """
    获取每日60秒早报

    Get the daily 60-second morning report

    :return:
    """
    try:
        res = requests.get(
            v2_url + "zaobao?format=json", headers={"token": alapi_token}
        )
        res = json.loads(res.text)

        if res["code"] != 200:
            return False, res["msg"]
        return True, res["data"]
    except Exception as e:
        return False, repr(e)


def doutu(keyword: str):
    """
    获取keyword的表情包

    :param keyword: 关键词
    :return: 图片链接
    """
    try:
        res = requests.post(
            v2_url + "doutu",
            params={"keyword": keyword},
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "token": alapi_token,
            },
        )
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"]
        return True, res["data"]
    except Exception as e:
        return False, repr(e)


def joke():
    """
    获取随机笑话

    :return: title, content
    """
    try:
        res = requests.get(v2_url + "joke/random", headers={"token": alapi_token})
        res = json.loads(res.text)
        if res["code"] != 200:
            return False, res["msg"]
        return True, res["data"]
    except Exception as e:
        return False, repr(e)



# coding=utf-8
"""
视频处理库, 你需要额外安装moviepy库

Video processing library, you need to install additional moviepy library
"""
import os
from .. import user_lang, qs_default_console, qs_error_string, requirePackage


def VideoWrapper(func):
    """
    视频处理的函数装饰器

    Video processing function decorator

    :param func: 函数
    :return: wrapper
    """

    def wrapper(path, *args, **kwargs):
        if not os.path.exists(path):
            qs_default_console.log(
                qs_error_string,
                f'{"No such file" if user_lang != "zh" else "文件不存在"}: {path}',
            )
            return
        if not os.path.isfile(path):
            qs_default_console.log(
                qs_error_string,
                f'{"No a file" if user_lang != "zh" else "不是文件"}: {path}',
            )
        mpy = requirePackage("moviepy")
        func(path, *args, **kwargs, mpy=mpy)

    return wrapper


@VideoWrapper
def video_2_gif(path: str, size: tuple = (480, 320), fps: int = None, mpy=None):
    """
    视频转gif（gif生成在视频路径的同目录下）

    Video transfer GIF (GIF generated in the same directory as the video path)

    :param mpy: moviepy.editor
    :param path: 视频路径
    :param size: gif尺寸
    :param fps: fps
    :return: None
    """
    file_name = ".".join(os.path.basename(path).split(".")[:-1]) + ".gif"
    dir_name = os.path.abspath(os.path.dirname(path))
    ct = mpy.VideoFileClip(path) if not size else mpy.VideoFileClip(path).resize(size)
    ct.write_gif(os.path.join(dir_name, file_name), fps=fps)


@VideoWrapper
def rm_audio(path: str, mpy=None):
    """
    删除视频的音频

    Delete the audio of the video

    :param mpy: moviepy.editor
    :param path: 视频路径
    :return: None
    """
    file_name = os.path.basename(path)
    if "." in file_name:
        indx = file_name.index(".")
        file_name = file_name[:indx] + "_rm_audio" + file_name[indx:]
        file_name = file_name.split(".")
        file_name = ".".join(file_name[:-1]) + ".mp4"
    else:
        file_name += "_rm_audio"
    dir_name = os.path.abspath(os.path.dirname(path))
    ct = mpy.VideoFileClip(path).set_audio(None)
    ct.write_videofile(os.path.join(dir_name, file_name))


@VideoWrapper
def tomp4(path: str, mpy=None):
    """
    将视频转为mp4

    Convert the video to MP4

    :param mpy: moviepy.editor
    :param path: 视频路径
    :return: None
    """
    file_name = os.path.basename(path)
    file_name = (
        ".".join(file_name.split(".")[:-1]) + ".mp4"
        if "." in file_name
        else file_name + ".mp4"
    )
    dir_name = os.path.abspath(os.path.dirname(path))
    ct = mpy.VideoFileClip(path)
    ct.write_videofile(os.path.join(dir_name, file_name))


@VideoWrapper
def video_2_mp3(path: str, mpy=None):
    """
    提取视频的音频并保存为mp3格式

    Extract the audio from the video and save it in MP3 format

    :return:
    """
    file_name = os.path.basename(path)
    file_name = (
        ".".join(file_name[:-1].split(".")) + ".mp3"
        if "." in file_name
        else file_name + ".mp3"
    )
    dir_name = os.path.abspath(os.path.dirname(path))
    mpy.VideoFileClip(path).audio.write_audiofile(os.path.join(dir_name, file_name))


@VideoWrapper
def set_video_audio(v_path: str, a_path: str, mpy=None):
    """
    设置视频的音频

    Set Video's Audio

    :param v_path: 视频路径 | video path
    :param a_path: 音频路径 | audio path
    :param mpy: None, DO NOT SET IT
    :return:
    """
    file_name = os.path.basename(v_path)
    file_name = (
        ".".join(file_name.split(".")[:-1]) + "_sta.mp4"
        if "." in file_name
        else file_name + "_sta.mp4"
    )
    dir_name = os.path.abspath(os.path.dirname(v_path))
    v = mpy.VideoFileClip(v_path)
    a = mpy.AudioFileClip(a_path)
    v = v.set_audio(a)
    v.write_videofile(os.path.join(dir_name, file_name), audio_codec="aac")



# coding=utf-8
"""
功能映射与菜单提示

Function mapping and menu prompt
"""
from . import qs_default_console, lang_detector
from .qsLesson import lesson

def menu_show(lines: list):
    from rich.table import Table, Column
    from rich.box import SIMPLE

    tb = Table(
        *(
            [
                Column("1", style="bold magenta", justify="left"),
                Column("2", style="bold yellow", justify="center"),
                Column("3", style="bold green", justify="right"),
            ]
        ),
        title='[bold underline]QuickStart_Rhy'
        + "\n",
        show_header=False,
        show_edge=False,
        box=SIMPLE,
    )

    for line in lines:
        tb.add_row('qs ' + line[0], lang_detector[line[0]] + lang_detector['tools_help'], line[1])
    qs_default_console.print(tb, justify="center")

def menu_table(group):
    from rich.table import Table, Column
    from rich.box import SIMPLE

    table = {
        'basic': basic_funcs,
        'system': system_funcs,
        'net': net_funcs,
        'api': api_funcs,
        'image': image_funcs,
    }[group]

    table.pop('self')
    tb = Table(
        *(
            [
                Column("1", style="bold magenta", justify="left"),
                Column("2", style="bold yellow", justify="right"),
            ]
        ),
        title=f'[bold underline]{group}'
        + "\n",
        show_header=False,
        show_edge=False,
        box=SIMPLE,
    )
    for key in table:
        if key in ['mktar', 'mkzip', 'mk7z']:
            tb.add_row(f"qs {key}", f"{lang_detector['compress'].format(key)}")
        elif key in ['unzip', 'untar', 'un7z', 'unrar']:
            tb.add_row(f"qs {key}", f"{lang_detector['decompress'].format(key)}")
        elif key in ['md5', 'sha1', 'sha256', 'sha512']:
            tb.add_row(f"qs {key}", f"{lang_detector['hashCal'].format(key)}")
        else:
            tb.add_row(f"qs {key}", f"{lang_detector['menu_' + key]}")
    qs_default_console.print(tb, justify="center")


def command_help(command):
    """
    """
    if val := cmd_table[command][command][1]:
        qs_default_console.print(f'[bold cyan][{lang_detector["usage"]}][/] qs {command} {val}')


basic_funcs = {
    "self": ".",
    "u": ["open_url", "<urls...>"],
    "a": ["open_app", "<app> [files...]"],
    "f": ["open_file", "<files...>"],
    "time": ["cur_time", ""],
    "cal": ["calculate", "<expression>"],
    "pcat": ["pcat", ""],
    "fcopy": ["fcopy", "<file>"],
    "tcopy": ["tcopy", "<text>"],
    "copy": ["copy", "<file>"],
    "play": ["play_music", "<audio>"],
    "lp": ["qs_print", "<file>"],
    "sas": ["sas", ""],
    "swap": ["swap", "<file1> <file2>"],
}

system_funcs = {
    "self": ".System",
    "top": ["top", ""],
    "go2git": ["go_github", ""],
    "mktar": ["mktar", "<files or dirs...>"],
    "untar": ["untar", "<tar file>"],
    "mkzip": ["mkzip", "<files or dirs...>"],
    "unzip": ["unzip", "<zip file>"],
    "unrar": ["unrar", "<rar file>"],
    "mk7z": ["mk7z", "<files or dirs...>"],
    "un7z": ["un7z", "<7z file>"],
    "md5": ["md5", "<file>"],
    "sha1": ["sha1", "<file>"],
    "sha256": ["sha256", "<file>"],
    "sha512": ["sha512", "<file>"],
    "diff": ["diff_dir", "<dir1> <dir2>"],
    "mount": ["mount_dmg", "<dmg file>"],
    "unmount": ["unmount_dmg", ""],
}

net_funcs = {
    "self": ".NetTools",
    "http": ["http", "[<ip>:<port>] [-bind <url>]"],
    "dl": ["download", f"""[url (in clipboard)]
      [--video] | [-v]    :-> {lang_detector['using_youtube_dl']}
      [--proxy] | [-px]   :-> {lang_detector['using_proxy']}
      [--name <fileName>] :-> {lang_detector['set_filename']}
      [--referer] | [-e] <url> :-> {lang_detector['set_referer']}
"""],
    "mdl": ["multi_download", "<urls separated by space>"],
    "wifi": ["wifi", ""],
    "upgrade": ["upgrade", ""],
    "netinfo": ["netinfo", "<ip | domain | url>"],
    "surl": ["scan_url", "<url>"],
}

api_funcs = {
    "self": ".ApiTools",
    "trans": ["translate", "<text (or in clipboard)>"],
    "pinyin": ["pinyin", "<text (or in clipboard)>"],
    "rmbg": ["remove_bg", "<image file>"],
    "smms": ["smms", "<image or markdown file>"],
    "alioss": ["ali_oss", f"""\
          -up <file> [bucket] :-> {lang_detector['upload_to_bucket']}
          -dl <file> [bucket] :-> {lang_detector['download_from_bucket']}
          -rm <file> [bucket] :-> {lang_detector['remove_from_bucket']}
          -ls [bucket]        :-> {lang_detector['list_bucket']}"""],
    "qiniu": ["qiniu", f"""\
          -up <file> [bucket] :-> {lang_detector['upload_to_bucket']}
          -dl <file> [bucket] :-> {lang_detector['download_from_bucket']}
          -rm <file> [bucket] :-> {lang_detector['remove_from_bucket']}
          -ls [bucket]        :-> {lang_detector['list_bucket']}"""],
    "txcos": ["txcos", f"""\
          -up <file> [bucket] :-> {lang_detector['upload_to_bucket']}
          -dl <file> [bucket] :-> {lang_detector['download_from_bucket']}
          -rm <file> [bucket] :-> {lang_detector['remove_from_bucket']}
          -ls [bucket]        :-> {lang_detector['list_bucket']}"""],
    "LG": ["largeImage", "<image file>"],
    "weather": ["weather", "[city]"],
    "nlp": ["AipNLP", "<text (or in clipboard)>"],
    "bcv": ["bili_cover", "<video id or url>"],
    "gbc": ["gbc", "<garbage name>"],
    "svi": ["short_video_info", "<url>"],
    "svd": ["short_video_dl", "<url>"],
    "acg": ["acg", "[--save]"],
    "bing": ["bingImg", "[--save]"],
    "phi": ["preview_html_images", "<url> [--save]"],
    "kd": ["kdCheck", "<courier id>"],
    "loli": ["loli", "[--save] [-p]"],
    "setu": ["setu", "[--save] [-p]"],
    "zhihu": ["zhihuDaily", ""],
    "wallhaven": ["wallhaven", "[--save] [-one] [--url <url>]"],
    "lmgtfy": ["lmgtfy", "<keywords>"],
    "d60": ["daily60s", "[--save]"],
    "m2t": ["m2t", "<magnet url>"],
    "d2m": ["d2m", "<keywords>"],
    "doutu": ["doutu", "<keywords>"],
    "joke": ["joke", ""],
    "gpt": ["gpt"],
    "ask": ['gpt_one'],
    "pushdeer": ["pushdeer", "[text]"],
}


image_funcs = {
    "self": ".ImageDeal",
    "stbg": ["set_img_background", "<image> <to color> \[from color]"],
    "icat": ["icat", "<image file or url>"],
    "v2gif": ["v2gif", "<video file> [size] [fps]"],
    "rmaudio": ["remove_audio", "<video file>"],
    "v2mp4": ["v2mp4", "<video file>"],
    "v2mp3": ["v2mp3", "<video file>"],
    "i2png": ["i2png", "<image file>"],
    "i2jpg": ["i2jpg", "<image file>"],
    "i2pdf": ["i2pdf", "<image file>"],
    "fmti": ["fmt_img_color", "<image file> <to color> [exclude colors...]"],
    "vsta": ["vsta", "<video file>"],
}


cmd_table = {
    j: i
    for i in [basic_funcs, api_funcs, net_funcs, image_funcs, system_funcs]
    for j in i
    if j != "self"
}



class GPTBotBase:
    def __init__(self, system_prompt, record_history=True):
        self.messages = [
            {
                "role": "system",
                "content": system_prompt,
            }
        ]
        self.record_history = record_history
    
    def save_conversation(self, file_path: str):
        with open(file_path, "w", encoding="utf-8") as f:
            for message in self.messages:
                f.write(f"## {message['role']}\n\n{message['content']}\n\n---\n")



# coding=utf-8
"""
qs的普通文件下载器！干翻迅雷，并且支持设置代理，设置referer，好看的命令行交互，啥j8文件都能下；(仅支持http[s]协议)

QS ordinary file downloader! Dry out Thunder, and support to set the proxy, set the Referer,
good-looking command line interaction, and it can download fuck any files; (only supports HTTP[S])

Author: RhythmLian (https://rhythmlian.cn)
"""
from . import size_format, get_fileinfo, headers
from concurrent.futures import ThreadPoolExecutor, wait
from ..SystemTools import get_core_num
from .. import (
    user_lang,
    qs_default_console,
    qs_default_status,
    qs_error_string,
    qs_info_string,
    qs_warning_string,
    requirePackage,
)
from threading import Lock
from requests import get
import signal
import queue
import time
import sys
import os

core_num = get_core_num()
maxBlockSize = int(5e6)
minBlockSize = int(5e5)


def GetBlockSize(sz):
    """
    自动规划块大小

    Automatically program the block size

    :param sz: 文件大小
    :return: 块大小
    """
    if sz >= maxBlockSize << 10:
        return maxBlockSize
    elif sz <= minBlockSize << 3:
        return minBlockSize
    else:
        return max(sz >> 9, minBlockSize)


def ask_overwrite(path: str):
    from .. import _ask

    return _ask(
        {
            "type": "confirm",
            "message": f'"{path}" 已存在, 是否覆盖?'
            if user_lang == "zh"
            else f'"{path}" already exists, overwrite?',
            "default": False,
        }
    )


class Downloader:
    from ..TuiTools.Bar import DataTransformBar

    def __init__(
        self,
        url: str,
        num: int,
        name: str = "",
        proxy: str = "",
        referer: str = "",
        output_error: bool = False,
        failed2exit: bool = False,
        exit_if_exist: bool = False,
        disableStatus: bool = False,
        disableParallel: bool = False,
        write_to_memory: bool = False,
        memory_stream_only: bool = False,
        ignore_404: bool = False,
    ):
        """
        qs普通文件下载引擎

        Qs general file download engine

        :param url: 文件url
        :param num: 线程数量
        :param name: 文件名
        :param proxy: 代理
        :param referer: referer
        :param output_error: 是否输出错误
        :param failed2exit: 是否下载失败就退出
        :param exit_if_exist: 如果文件已存在，是否退出
        :param disableStatus: 是否禁用状态栏
        :param disableParallel: 是否禁用并行下载
        :param write_to_memory: 是否写入内存
        :param ignore_404: 是否忽略404错误
        """
        signal.signal(signal.SIGINT, self._kill_self)
        info_flag = True
        self.url, self.num, self.output_error, self.proxies = url, num, output_error, {}
        self.exit_if_exist = exit_if_exist
        self.enabled = True
        self.disableStatus = disableStatus
        self.write_to_memory = write_to_memory
        self.memory_stream_only = memory_stream_only
        if not self.disableStatus:
            with qs_default_status(
                "Getting file info.." if user_lang != "zh" else "获取文件信息中.."
            ):
                self.url, self.name, r = get_fileinfo(url, proxy, referer)
        else:
            self.url, self.name, r = get_fileinfo(url, proxy, referer)
        if not (self.url and self.name and r):
            info_flag = False
            if not self.disableStatus:
                qs_default_console.print(
                    qs_error_string if failed2exit else qs_warning_string,
                    "Get File information failed, please check network!"
                    if user_lang != "zh"
                    else "获取文件信息失败，请检查网络!",
                )
            if failed2exit and not ignore_404:
                self.enabled = False
                return
        if not self.url:
            qs_default_console.print(
                qs_error_string,
                "Connection Error!" if user_lang != "zh" else "连接失败!",
                self.name,
            )
            self.enabled = False
            return

        if self.name and "." not in self.name:
            self.name = os.path.basename(url)
        if name:
            self.name = name

        if os.path.exists(self.name) and self.exit_if_exist:
            self.enabled = False
            return

        if proxy:
            self.proxies = {"http": "http://" + proxy, "https": "https://" + proxy}
        self.headers = headers
        if referer:
            self.headers["Referer"] = referer
        try:
            if not info_flag:
                raise KeyError
            self.size = int(r.headers["content-length"])
            self.main_progress = Downloader.DataTransformBar()
            self.dl_id = self.main_progress.add_task(
                "Download", filename=self.name, start=False
            )
            self.main_progress.update(self.dl_id, total=self.size)
            if self.size < 5e6 or disableParallel:
                if not self.disableStatus:
                    qs_default_console.print(
                        qs_info_string,
                        "FILE SIZE" if user_lang != "zh" else "文件大小",
                        size_format(self.size),
                    )
                self.size = -self.size
            else:
                self.fileBlock = GetBlockSize(self.size)
        except KeyError:
            self.size = -1
            self.main_progress = Downloader.DataTransformBar(False)
            self.dl_id = self.main_progress.add_task(
                "Download", filename=self.name, start=False
            )
        if self.size > 0:
            self.pool = ThreadPoolExecutor(max_workers=self.num)
            self.futures, self.fileLock = [], Lock()
            self.job_queue = queue.Queue()
            if os.path.exists(self.name + ".qs_dl"):
                qs_default_console.print(
                    qs_info_string,
                    "Resuming task from:" if user_lang != "zh" else "从文件恢复任务:",
                    self.name + ".qs_dl",
                )
                self.ctn_file = open(self.name + ".qs_dl", "r+")
                self.ctn = {int(i) for i in self.ctn_file.read().strip().split()}
            elif not self.write_to_memory:
                self.ctn_file = open(self.name + ".qs_dl", "w")
                self.ctn = set()
            else:
                self.ctn = set()
            self.writers = (
                requirePackage(".ThreadTools", "FileWriters")(
                    self.name, max(2, int(core_num / 2)), "rb+" if self.ctn else "wb"
                )
                if not self.write_to_memory
                else requirePackage(".ThreadTools", "MemWriter")()
            )
            if not self.disableStatus:
                qs_default_console.print(
                    qs_info_string,
                    "FILE  SIZE" if user_lang != "zh" else "文件大小",
                    size_format(self.size, align=True),
                )
                qs_default_console.print(
                    qs_info_string,
                    "THRAED NUM" if user_lang != "zh" else "线程数量",
                    "%7d" % num,
                )

    def _kill_self(self, signum, frame):
        """
        终止下载任务，保存断点

        Terminate the download and save the breakpoint

        :param signum: 信号
        :param frame: frame
        :return: None
        """
        if self.size > 0:
            self.main_progress.stop_task(self.dl_id)
            self.main_progress.stop()
            qs_default_console.print(
                qs_info_string,
                "Get Ctrl-C, exiting..." if user_lang != "zh" else "捕获Ctrl-C, 正在退出...",
            )
            if not self.write_to_memory:
                self.ctn_file.close()
            self.pool.shutdown(wait=False)
            self.writers.wait()
            qs_default_console.print(
                qs_info_string, "Deal Done!" if user_lang != "zh" else "处理完成!"
            )
        os._exit(0)

    def _dl(self, start):
        """
        执行文件块下载任务

        Perform the file block download task

        :param start: 文件块起始偏移量
        :return: None
        """
        try:
            _sz = min(start + self.fileBlock, self.size - 1)
            _headers = self.headers.copy()
            _headers["Range"] = "bytes={}-{}".format(start, _sz)
            _content = b""
            for chunk in get(
                self.url, headers=_headers, timeout=50, proxies=self.proxies
            ).iter_content(65536):
                _content += chunk
                if not self.disableStatus:
                    self.main_progress.advance(self.dl_id, sys.getsizeof(chunk))
            self.writers.new_job(_content, start)
        except Exception as e:
            if self.output_error:
                qs_default_console.print(qs_error_string, repr(e))
            self.job_queue.put(start)
        else:
            self.ctn.add(start)
            if not self.write_to_memory:
                self.fileLock.acquire()
                self.ctn_file.write("%d\n" % start)
                self.fileLock.release()

    def _single_dl(self):
        """
        无法并行下载或无需并行下载时采用串行下载

        Parallel downloads are not possible or serial downloads are not required

        :return: None
        """
        r = get(self.url, stream=True, proxies=self.proxies, headers=self.headers)
        flag = self.size != -1

        if not self.disableStatus:
            if flag:
                self.main_progress.start_task(self.dl_id)
            else:
                self.main_progress.update(self.dl_id, total=-1)
        if self.write_to_memory == False:
            with open(self.name, "wb") as f:
                for chunk in r.iter_content(32768):
                    f.write(chunk)
                    self.main_progress.advance(self.dl_id, sys.getsizeof(chunk))
        else:
            if self.memory_stream_only:
                self.name = r, self.size
            self.name = b""
            for chunk in r.iter_content(32768):
                self.name += chunk
                self.main_progress.advance(self.dl_id, sys.getsizeof(chunk))

    def run(self):
        """
        规划下载任务并开始下载

        qs可以有效应对网络问题对下载任务造成的影响，如：
          1.下载过程中切换代理

          2.链路异常
        并确保下载任务顺利完成

        :return: None
        """
        if not self.enabled:
            return self.name if self.exit_if_exist else None
        if not self.disableStatus:
            self.main_progress.start()
        if self.size > 0:
            if not self.disableStatus:
                self.main_progress.start_task(self.dl_id)
            if not self.write_to_memory:
                if not self.ctn:
                    with open(self.name, "wb") as fp:
                        fp.truncate(self.size)
                if not self.disableStatus:
                    self.main_progress.advance(
                        self.dl_id, self.fileBlock * len(self.ctn)
                    )
                for i in range(0, self.size, self.fileBlock):
                    if self.ctn and i in self.ctn:
                        continue
                    else:
                        self.job_queue.put(i)
            elif self.memory_stream_only:
                self._single_dl()
            else:
                for i in range(0, self.size, self.fileBlock):
                    self.job_queue.put(i)
            retry_cnt = 0
            while not self.job_queue.empty():
                self.futures.clear()
                while not self.job_queue.empty():
                    cur = self.job_queue.get()
                    if cur not in self.ctn:
                        self.futures.append(self.pool.submit(self._dl, cur))
                wait(self.futures)
                if not self.job_queue.empty() and retry_cnt > 2:
                    qs_default_console.print(
                        qs_warning_string,
                        "Exists File Block Lost, Retrying after 0.5 sec"
                        if user_lang != "zh"
                        else "存在文件块丢失，0.5秒后重试",
                    )
                    time.sleep(0.5)
                retry_cnt += 1
            self.writers.wait()
            if not self.write_to_memory:
                self.ctn_file.close()
                os.remove(self.name + ".qs_dl")
            elif not self.memory_stream_only:
                self.name = self.writers.content
        else:
            self._single_dl()
        if not self.disableStatus:
            self.main_progress.stop()
            if self.write_to_memory or self.memory_stream_only:
                qs_default_console.print(
                    qs_info_string,
                    "download done!" if user_lang != "zh" else "下载完成!",
                )
            else:
                qs_default_console.print(
                    qs_info_string,
                    self.name,
                    "download done!" if user_lang != "zh" else "下载完成!",
                )
        return self.name


def normal_dl(
    url,
    set_name: str = "",
    set_proxy: str = "",
    set_referer: str = "",
    thread_num: int = min(16, core_num * 4),
    output_error: bool = False,
    failed2exit: bool = False,
    exit_if_exist: bool = False,
    disableStatus: bool = False,
    disableParallel: bool = False,
    write_to_memory: bool = False,
    memory_stream_only: bool = False,
    ignore_404: bool = False,
):
    """
    自动规划下载线程数量并开始并行下载

    Automatically schedule the number of download threads and begin parallel downloads

    :param disableStatus:
    :param url: 文件url
    :param set_name: 设置文件路径
    :param set_proxy: 设置代理（默认无代理）
    :param set_referer: 设置referer
    :param thread_num: 线程数
    :param output_error: 输出报错信息
    :param failed2exit: 获取文件信息失败则不下载, 否则qs将继续尝试下载
    :param exit_if_exist: 如果文件已存在则不下载
    :param disableStatus: 是否显示当前任务状态
    :param disableParallel: 是否禁用并行下载
    :param write_to_memory: 是否将下载内容直接写入内存（默认写入文件）
    :param ignore_404: 忽略404错误, 因无法获取文件信息，所以需自行设置文件名
    :return: 文件路径或二进制内容 | file path or bytes
    """
    if set_name and os.path.exists(set_name) and not ask_overwrite(set_name):
        return ""
    return Downloader(
        url=url,
        num=thread_num,
        name=set_name,
        proxy=set_proxy,
        referer=set_referer,
        output_error=output_error,
        failed2exit=failed2exit,
        exit_if_exist=exit_if_exist,
        disableStatus=disableStatus,
        disableParallel=disableParallel,
        write_to_memory=write_to_memory,
        memory_stream_only=memory_stream_only,
        ignore_404=ignore_404,
    ).run()
